# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import uuid

from oslo_serialization import jsonutils
import requests

from glare.tests.functional import base


def sort_results(lst, target='name'):
    return sorted(lst, key=lambda x: x[target])


class TestList(base.TestArtifact):
    def test_list_marker_and_limit(self):
        # Create artifacts
        art_list = [self.create_artifact({'name': 'name%s' % i,
                                          'version': '1.0',
                                          'tags': ['tag%s' % i],
                                          'int1': 1024 + i,
                                          'float1': 123.456,
                                          'str1': 'bugaga',
                                          'bool1': True})
                    for i in range(5)]

        # sort by 'next' url
        url = '/sample_artifact?limit=1&sort=int1:asc,name:desc'
        result = self.get(url=url)
        self.assertEqual([art_list[0]], result['sample_artifact'])
        marker = result['next']
        result = self.get(url=marker[10:])
        self.assertEqual([art_list[1]], result['sample_artifact'])

        # sort by custom marker
        url = '/sample_artifact?sort=int1:asc&marker=%s' % art_list[1]['id']
        result = self.get(url=url)
        self.assertEqual(art_list[2:], result['sample_artifact'])
        url = '/sample_artifact?sort=int1:desc&marker=%s' % art_list[1]['id']
        result = self.get(url=url)
        self.assertEqual(art_list[:1], result['sample_artifact'])
        url = '/sample_artifact' \
              '?sort=float1:asc,name:desc&marker=%s' % art_list[1]['id']
        result = self.get(url=url)
        self.assertEqual([art_list[0]], result['sample_artifact'])

        # paginate by name in desc order with limit 2
        url = '/sample_artifact?limit=2&sort=name:desc'
        result = self.get(url=url)
        self.assertEqual(art_list[4:2:-1], result['sample_artifact'])
        marker = result['next']
        result = self.get(url=marker[10:])
        self.assertEqual(art_list[2:0:-1], result['sample_artifact'])
        marker = result['next']
        result = self.get(url=marker[10:])
        self.assertEqual([art_list[0]], result['sample_artifact'])

    def test_list_base_filters(self):
        # Create artifact
        art_list = [self.create_artifact({'name': 'name%s' % i,
                                          'version': '1.0',
                                          'tags': ['tag%s' % i],
                                          'int1': 1024,
                                          'float1': 123.456,
                                          'str1': 'bugaga',
                                          'bool1': True})
                    for i in range(5)]

        public_art = self.create_artifact({'name': 'name5',
                                           'version': '1.0',
                                           'tags': ['tag4', 'tag5'],
                                           'int1': 2048,
                                           'float1': 987.654,
                                           'str1': 'lalala',
                                           'bool1': False,
                                           'string_required': '123'})
        url = '/sample_artifact/%s' % public_art['id']
        data = [{
            "op": "replace",
            "path": "/status",
            "value": "active"
        }]
        self.patch(url=url, data=data, status=200)
        public_art = self.admin_action(public_art['id'], self.make_public)

        art_list.append(public_art)

        art_list.sort(key=lambda x: x['name'])

        url = '/sample_artifact?str1=bla:empty'
        self.get(url=url, status=400)

        url = '/sample_artifact?str1=bla:empty'
        self.get(url=url, status=400)

        url = '/sample_artifact?name=name0'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([art_list[0]], result)

        url = '/sample_artifact?tags=tag4'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[4:], result)

        url = '/sample_artifact?name=eq:name0'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:1], result)

        url = '/sample_artifact?str1=eq:bugaga'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?int1=eq:2048'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?float1=eq:123.456'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?name=neq:name0'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[1:], result)

        url = '/sample_artifact?name=in:name,name0'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:1], result)

        url = '/sample_artifact?name=in:not_exist,name0'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:1], result)

        url = '/sample_artifact?name=not_exist'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual([], result)

        url = '/sample_artifact?name=bla:name1'
        self.get(url=url, status=400)

        url = '/sample_artifact?name='
        self.get(url=url, status=400)

        url = '/sample_artifact?name=eq:'
        self.get(url=url, status=400)

        url = '/sample_artifact?tags=tag4,tag5'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?tags-any=tag4'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[4:], result)

        url = '/sample_artifact?tags=tag4,tag_not_exist,tag5'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual([], result)

        url = '/sample_artifact?tags-any=tag4,tag_not_exist,tag5'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[4:], result)

        url = '/sample_artifact?tags=tag_not_exist,tag_not_exist_1'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual([], result)

        url = '/sample_artifact?tags'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list, result)

        url = '/sample_artifact?tags='
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list, result)

        url = '/sample_artifact?tags=eq:tag0'
        self.get(url=url, status=400)

        url = '/sample_artifact?tags=bla:tag0'
        self.get(url=url, status=400)

        url = '/sample_artifact?tags=neq:tag1'
        self.get(url=url, status=400)

        url = '/sample_artifact?visibility=private'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?visibility=public'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?visibility=eq:private'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?visibility=eq:public'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?visibility=neq:private'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?visibility=neq:public'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?visibility=blabla'
        self.get(url=url, status=400)

        url = '/sample_artifact?visibility=neq:blabla'
        self.get(url=url, status=400)

        url = '/sample_artifact?name=eq:name0&name=name1&tags=tag1'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual([], result)

        url = '/sample_artifact?int1=gt:2000'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?int1=lte:1024'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?int1=gt:1000&int1=lt:2000'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?int1=lt:2000'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?float1=gt:200.000'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

        url = '/sample_artifact?float1=gt:100.00&float1=lt:200.00'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?float1=lt:200.00'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?float1=lt:200'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?float1=lte:123.456'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?bool1=True'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:5], result)

        url = '/sample_artifact?bool1=False'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[5:], result)

    def test_artifact_list_dict_filters(self):
        lists_of_str = [
            ['aaa', 'bbb', 'ccc'],
            ['aaa', 'bbb'],
            ['aaa', 'ddd'],
            ['bbb'],
            ['ccc']
        ]
        dicts_of_str = [
            {'aaa': 'z', 'bbb': 'z', 'ccc': 'z'},
            {'aaa': 'z', 'bbb': 'z'},
            {'aaa': 'z', 'ddd': 'z'},
            {'bbb': 'z'},
            {'ccc': 'z'}
        ]
        art_list = [self.create_artifact({'name': 'name%s' % i,
                                          'version': '1.0',
                                          'tags': ['tag%s' % i],
                                          'int1': 1024,
                                          'float1': 123.456,
                                          'str1': 'bugaga',
                                          'bool1': True,
                                          'list_of_str': lists_of_str[i],
                                          'dict_of_str': dicts_of_str[i]})
                    for i in range(5)]

        # test list filters
        url = '/sample_artifact?list_of_str=aaa&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:3], result)

        url = '/sample_artifact?list_of_str=ccc&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([art_list[0], art_list[4]], result)

        url = '/sample_artifact?list_of_str=eee&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        # test dict filters
        url = '/sample_artifact?dict_of_str=aaa&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:3], result)

        url = '/sample_artifact?dict_of_str=ccc&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([art_list[0], art_list[4]], result)

        url = '/sample_artifact?dict_of_str=eee&sort=name'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

    def test_list_dict_prop_filters(self):
        # Create artifact
        art_list = [self.create_artifact({'name': 'name0',
                                          'version': '1.0',
                                          'dict_of_str': {'pr1': 'val1'}}),
                    self.create_artifact({'name': 'name1',
                                          'version': '1.0',
                                          'dict_of_str': {'pr1': 'val1',
                                                          'pr2': 'val2'}}),
                    self.create_artifact({'name': 'name2',
                                          'version': '1.0',
                                          'dict_of_str': {'pr3': 'val3'}}),
                    self.create_artifact({'name': 'name3',
                                          'version': '1.0',
                                          'dict_of_str': {'pr3': 'val1'},
                                          'dict_of_int': {"1": 10, "2": 20}}),
                    self.create_artifact({'name': 'name4',
                                          'version': '1.0',
                                          'dict_of_str': {},
                                          'dict_of_int': {"2": 20, "3": 30}}),
                    ]

        art_list.sort(key=lambda x: x['name'])

        url = '/sample_artifact?dict_of_str.pr1=val1'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:2], result)

        url = '/sample_artifact?dict_of_int.1=10'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[3:4], result)

        url = '/sample_artifact?dict_of_str.pr1=val999'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        url = '/sample_artifact?dict_of_str.pr1=eq:val1'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual(art_list[:2], result)

        url = '/sample_artifact?dict_of_str.'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        for op in ['gt', 'gte', 'lt', 'lte']:
            url = '/sample_artifact?dict_of_str.pr3=%s:val3' % op
            self.get(url=url, status=400)

        url = '/sample_artifact?dict_of_str.pr3=blabla:val3'
        self.get(url=url, status=400)

        url = '/sample_artifact?dict_of_str.pr1='
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        url = '/sample_artifact?dict_of_str.pr1='
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        url = '/sample_artifact?dict_of_str'
        self.assertEqual([], result)

        url = '/sample_artifact?dict_of_str.pr3=blabla:val3'
        self.get(url=url, status=400)

        url = '/sample_artifact?list_of_str.pr3=blabla:val3'
        self.get(url=url, status=400)

        url = '/sample_artifact?dict_of_str.bla=val1'
        result = sort_results(self.get(url=url)['sample_artifact'])
        self.assertEqual([], result)

        url = '/sample_artifact?dict_of_int.1=lala'
        self.get(url=url, status=400)

    def test_list_sorted(self):
        art_list = [self.create_artifact({'name': 'name%s' % i,
                                          'version': '1.0',
                                          'tags': ['tag%s' % i],
                                          'int1': i,
                                          'float1': 123.456 + (-0.9) ** i,
                                          'str1': 'bugaga',
                                          'bool1': True,
                                          'list_of_int': [11, 22, - i],
                                          'dict_of_int': {'one': 4 * i,
                                                          'two': (-2) ** i}})
                    for i in range(5)]

        # sorted by string 'asc'
        url = '/sample_artifact?sort=name:asc'
        result = self.get(url=url)
        expected = sort_results(art_list)
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by string 'desc'
        url = '/sample_artifact?sort=name:desc'
        result = self.get(url=url)
        expected = sort_results(art_list)
        expected.reverse()
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by int 'asc'
        url = '/sample_artifact?sort=int1:asc'
        result = self.get(url=url)
        expected = sort_results(art_list, target='int1')
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by int 'desc'
        url = '/sample_artifact?sort=int1:desc'
        result = self.get(url=url)
        expected = sort_results(art_list, target='int1')
        expected.reverse()
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by float 'asc'
        url = '/sample_artifact?sort=float1:asc'
        result = self.get(url=url)
        expected = sort_results(art_list, target='float1')
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by float 'desc'
        url = '/sample_artifact?sort=float1:desc'
        result = self.get(url=url)
        expected = sort_results(art_list, target='float1')
        expected.reverse()
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by unsorted 'asc'
        url = '/sample_artifact?sort=bool1:asc'
        self.get(url=url, status=400)

        # sorted by unsorted 'desc'
        url = '/sample_artifact?sort=bool1:desc'
        self.get(url=url, status=400)

        # sorted by non-existent 'asc'
        url = '/sample_artifact?sort=non_existent:asc'
        self.get(url=url, status=400)

        # sorted by non-existent 'desc'
        url = '/sample_artifact?sort=non_existent:desc'
        self.get(url=url, status=400)

        # sorted by invalid op
        url = '/sample_artifact?sort=name:invalid_op'
        self.get(url=url, status=400)

        # sorted without op
        url = '/sample_artifact?sort=name'
        result = self.get(url=url)
        expected = sort_results(art_list)
        expected.reverse()
        self.assertEqual(expected, result['sample_artifact'])

        # sorted by list
        url = '/sample_artifact?sort=list_of_int:asc'
        self.get(url=url, status=400)

        # sorted by dict
        url = '/sample_artifact?sort=dict_of_int:asc'
        self.get(url=url, status=400)

        # sorted by element of dict
        url = '/sample_artifact?sort=dict_of_int.one:asc'
        self.get(url=url, status=400)

        # sorted by any prop
        url = '/sample_artifact?sort=name:asc,int1:desc'
        result = self.get(url=url)
        expected = sort_results(sort_results(art_list), target='int1')
        self.assertEqual(expected, result['sample_artifact'])

    def test_list_versions(self):
        # Create artifacts with versions
        version_list = ['1.0', '1.1', '2.0.0', '2.0.1-beta', '2.0.1', '20.0']

        # Create artifact
        art_list = [self.create_artifact({'name': 'name',
                                          'version': version_list[i - 1],
                                          'tags': ['tag%s' % i],
                                          'int1': 2048,
                                          'float1': 123.456,
                                          'str1': 'bugaga',
                                          'bool1': True})
                    for i in range(1, 7)]

        public_art = self.create_artifact(
            {'name': 'name',
             'tags': ['tag4', 'tag5'],
             'int1': 1024,
             'float1': 987.654,
             'str1': 'lalala',
             'bool1': False,
             'string_required': '123'})
        url = '/sample_artifact/%s' % public_art['id']
        data = [{
            "op": "replace",
            "path": "/status",
            "value": "active"
        }]
        self.patch(url=url, data=data, status=200)
        public_art = self.admin_action(public_art['id'], self.make_public)

        art_list.insert(0, public_art)

        expected_result = sort_results(art_list, target='version')
        url = '/sample_artifact'
        result = sort_results(self.get(url=url)['sample_artifact'],
                              target='version')
        self.assertEqual(expected_result, result)

        # Creating an artifact with existing version fails
        self.create_artifact(
            {'name': 'name',
             'version': '1.0',
             'tags': ['tag1'],
             'int1': 2048,
             'float1': 123.456,
             'str1': 'bugaga',
             'bool1': True},
            status=409)

        url = '/sample_artifact?name=name&version=gte:2.0.0'
        result = sort_results(self.get(url=url)['sample_artifact'],
                              target='version')
        self.assertEqual(expected_result[3:], result)

        url = ('/sample_artifact?'
               'name=name&version=gte:1.1&version=lt:2.0.1-beta')
        result = sort_results(self.get(url=url)['sample_artifact'],
                              target='version')
        self.assertEqual(expected_result[2:4], result)

        # Filtering by version without name is ok
        url = '/sample_artifact?version=gte:2.0.0'
        self.get(url=url, status=200)

        # Several name filters with version is ok
        url = '/sample_artifact?name=name&name=anothername&version=gte:2.0.0'
        self.get(url=url, status=200)

        # Filtering by version with name filter op different from 'eq'
        url = '/sample_artifact?version=gte:2.0.0&name=neq:name'
        self.get(url=url, status=200)

        # Sorting by version 'asc'
        url = '/sample_artifact?name=name&sort=version:asc'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual(art_list, result)

        # Sorting by version 'desc'
        url = '/sample_artifact?name=name&sort=version:desc'
        result = self.get(url=url)['sample_artifact']
        self.assertEqual(list(reversed(art_list)), result)

    def test_list_latest_filter(self):
        # Create artifacts with versions
        group1_versions = ['1.0', '20.0', '2.0.0', '2.0.1-beta', '2.0.1']
        group2_versions = ['1', '1000.0.1-beta', '99.0',
                           '1000.0.1-alpha', '1000.0.1']

        for i in range(5):
            self.create_artifact(
                {'name': 'group1',
                 'version': group1_versions[i],
                 'tags': ['tag%s' % i],
                 'int1': 2048 + i,
                 'float1': 123.456,
                 'str1': 'bugaga',
                 "string_required": "test_str",
                 'bool1': True})
            self.create_artifact(
                {'name': 'group2',
                 'version': group2_versions[i],
                 'tags': ['tag%s' % i],
                 'int1': 2048 + i,
                 'float1': 123.456,
                 'str1': 'bugaga',
                 "string_required": "test_str",
                 'bool1': True})

        url = '/sample_artifact?version=latest&sort=name:asc'
        res = self.get(url=url, status=200)['sample_artifact']
        self.assertEqual(2, len(res))
        self.assertEqual('20.0.0', res[0]['version'])
        self.assertEqual('1000.0.1', res[1]['version'])

        self.patch('/sample_artifact/' + res[0]['id'], self.make_active)

        url = '/sample_artifact?version=latest&sort=name:asc&status=drafted'
        res = self.get(url=url, status=200)['sample_artifact']
        self.assertEqual(2, len(res))
        self.assertEqual('2.0.1', res[0]['version'])
        self.assertEqual('1000.0.1', res[1]['version'])

        url = '/sample_artifact?version=latest&sort=name:asc&int1=2050'
        res = self.get(url=url, status=200)['sample_artifact']
        self.assertEqual(2, len(res))
        self.assertEqual('2.0.0', res[0]['version'])
        self.assertEqual('99.0.0', res[1]['version'])

        url = '/sample_artifact?version=latest&name=group1'
        res = self.get(url=url, status=200)['sample_artifact']
        self.assertEqual(1, len(res))
        self.assertEqual('20.0.0', res[0]['version'])

        url = '/sample_artifact?version=latest&name=group2'
        res = self.get(url=url, status=200)['sample_artifact']
        self.assertEqual(1, len(res))
        self.assertEqual('1000.0.1', res[0]['version'])

    def test_list_support_unicode_filters(self):
        unicode_text = u'\u041f\u0420\u0418\u0412\u0415\u0422'
        art1 = self.create_artifact(data={'name': unicode_text})
        self.assertEqual(unicode_text, art1['name'])

        mixed_text = u'la\u041f'
        art2 = self.create_artifact(data={'name': mixed_text})
        self.assertEqual(mixed_text, art2['name'])

        headers = {'Content-Type': 'text/html; charset=UTF-8'}
        url = u'/sample_artifact?name=\u041f\u0420\u0418\u0412\u0415\u0422'
        response_url = u'/artifacts/sample_artifact?name=' \
                       u'%D0%9F%D0%A0%D0%98%D0%92%D0%95%D0%A2'
        result = self.get(url=url, headers=headers)
        self.assertEqual(art1, result['sample_artifact'][0])
        self.assertEqual(response_url, result['first'])


class TestBlobs(base.TestArtifact):
    def test_blob_dicts(self):
        # Getting empty artifact list
        url = '/sample_artifact'
        response = self.get(url=url, status=200)
        expected = {'first': '/artifacts/sample_artifact',
                    'sample_artifact': [],
                    'schema': '/schemas/sample_artifact'}
        self.assertEqual(expected, response)

        # Create a test artifact
        art = self.create_artifact(status=201,
                                   data={'name': 'test',
                                         'version': '1.0',
                                         'string_required': '123'})
        self.assertIsNotNone(art['id'])

        # Get the artifact which should have a generated id and status
        # 'drafted'
        url = '/sample_artifact/%s' % art['id']
        art_1 = self.get(url=url, status=200)
        self.assertIsNotNone(art_1['id'])
        self.assertEqual('drafted', art_1['status'])

        # Upload data to blob dict
        headers = {'Content-Type': 'application/octet-stream'}
        data = "data" * 100
        blob_name = 'blob_name' * 100
        self.put(url=url + '/dict_of_blobs/' + blob_name,
                 data=data, status=200, headers=headers)

        # Download data from blob dict
        self.assertEqual(data,
                         self.get(url=url + '/dict_of_blobs/' + blob_name,
                                  status=200))

        # Download blob from undefined dict property
        self.get(url=url + '/not_a_dict/not_a_blob', status=400)

        # Blob url is generated right
        art = self.get(url=url, status=200)
        exp_blob_url = '/artifacts' + url + '/dict_of_blobs/' + blob_name
        self.assertEqual(exp_blob_url,
                         art['dict_of_blobs'][blob_name]['url'])

    def test_blob_upload(self):
        # create artifact with blob
        data = 'data'
        self.create_artifact(
            data={'name': 'test_af', 'blob': data,
                  'version': '0.0.1'}, status=400)
        art = self.create_artifact(data={'name': 'test_af',
                                         'version': '0.0.1',
                                         'string_required': 'test'})
        url = '/sample_artifact/%s' % art['id']
        headers = {'Content-Type': 'application/octet-stream',
                   'Content-Length': '4'}

        # upload to non-existing property
        self.put(url=url + '/blob_non_exist', data=data, status=400,
                 headers=headers)

        # upload too big value
        big_data = "this is the smallest big data"
        self.put(url=url + '/small_blob', data=big_data, status=413,
                 headers=headers)
        # upload correct blob value
        self.put(url=url + '/small_blob', data=big_data[:2], headers=headers)

        # Upload artifact via different user
        self.set_user('user2')
        self.put(url=url + '/blob', data=data, status=404,
                 headers=headers)

        # Upload file to the artifact
        self.set_user('user1')
        art = self.put(url=url + '/blob', data=data, status=200,
                       headers=headers)
        self.assertEqual('active', art['blob']['status'])
        self.assertEqual('application/octet-stream',
                         art['blob']['content_type'])
        self.assertIn('url', art['blob'])
        self.assertNotIn('id', art['blob'])

        # Blob url is generated right
        exp_blob_url = '/artifacts' + url + '/blob'
        self.assertEqual(exp_blob_url, art['blob']['url'])

        # reUpload file to artifact
        self.put(url=url + '/blob', data=data, status=409,
                 headers=headers)
        # upload blob dict
        self.put(url + '/dict_of_blobs/test_key', data=data, headers=headers)
        # test re-upload failed
        self.put(url + '/dict_of_blobs/test_key', data=data, headers=headers,
                 status=409)

        # upload few other blobs to the dict
        for elem in ('aaa', 'bbb', 'ccc', 'ddd'):
            self.put(url + '/dict_of_blobs/' + elem, data=data,
                     headers=headers)

        # upload to active artifact
        self.patch(url, self.make_active)
        self.put(url + '/dict_of_blobs/key2', data=data, status=403,
                 headers=headers)

        self.delete(url)

    def test_blob_download(self):
        data = 'some_arbitrary_testing_data'
        art = self.create_artifact(data={'name': 'test_af',
                                         'version': '0.0.1'})
        url = '/sample_artifact/%s' % art['id']

        # download not uploaded blob
        self.get(url=url + '/blob', status=404)

        # download blob from not existing artifact
        self.get(url=url + '1/blob', status=404)

        # download blob from undefined property
        self.get(url=url + '/not_a_blob', status=400)

        headers = {'Content-Type': 'application/octet-stream'}
        art = self.put(url=url + '/blob', data=data, status=200,
                       headers=headers)
        self.assertEqual('active', art['blob']['status'])
        md5 = hashlib.md5(data.encode('UTF-8')).hexdigest()
        sha1 = hashlib.sha1(data.encode('UTF-8')).hexdigest()
        sha256 = hashlib.sha256(data.encode('UTF-8')).hexdigest()
        self.assertEqual(md5, art['blob']['md5'])
        self.assertEqual(sha1, art['blob']['sha1'])
        self.assertEqual(sha256, art['blob']['sha256'])

        # check that content-length is in response
        response = requests.get(self._url(url + '/blob'),
                                headers=self._headers())
        self.assertEqual('27', response.headers["content-length"])

        # check that all checksums are in response
        response = requests.get(self._url(url + '/blob'),
                                headers=self._headers())
        self.assertEqual('0825587cc011b7e76381b65e19d5ec27',
                         response.headers["Content-MD5"])
        self.assertEqual('89eb4b969b721ba8c3aff18ad7d69454f651a697',
                         response.headers["X-Openstack-Glare-Content-SHA1"])
        self.assertEqual('bbfd48c7ec792fc462e58232d4d9f407'
                         'ecefb75cc9e9823336166556b499ea4d',
                         response.headers["X-Openstack-Glare-Content-SHA256"])

        blob_data = self.get(url=url + '/blob')
        self.assertEqual(data, blob_data)

        # download artifact via admin
        self.set_user('admin')
        blob_data = self.get(url=url + '/blob')
        self.assertEqual(data, blob_data)

        # try to download blob via different user
        self.set_user('user2')
        self.get(url=url + '/blob', status=404)

    def test_blob_add_custom_location(self):
        # Create artifact
        art = self.create_artifact({'name': 'name5',
                                    'version': '1.0',
                                    'tags': ['tag1', 'tag2', 'tag3'],
                                    'int1': 2048,
                                    'float1': 987.654,
                                    'str1': 'lalala',
                                    'bool1': False,
                                    'string_required': '123'})
        self.assertIsNotNone(art['id'])

        # Set custom location
        url = '/sample_artifact/%s' % art['id']
        body = jsonutils.dumps(
            {'url': 'https://www.apache.org/licenses/LICENSE-2.0.txt',
             'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"})
        headers = {'Content-Type':
                   'application/vnd+openstack.glare-custom-location+json'}
        self.put(url=url + '/blob', data=body,
                 status=200, headers=headers)

        # test re-add failed
        self.put(url=url + '/blob', data=body, status=409, headers=headers)
        # add to non-existing property
        self.put(url=url + '/blob_non_exist', data=body, status=400,
                 headers=headers)

        # Get the artifact, blob property should have status 'active'
        art = self.get(url=url, status=200)
        self.assertEqual('active', art['blob']['status'])
        self.assertEqual('fake', art['blob']['md5'])
        self.assertEqual('fake_sha', art['blob']['sha1'])
        self.assertEqual('fake_sha256', art['blob']['sha256'])
        self.assertIsNone(art['blob']['size'])
        self.assertIsNone(art['blob']['content_type'])
        self.assertEqual('https://www.apache.org/licenses/LICENSE-2.0.txt',
                         art['blob']['url'])
        self.assertNotIn('id', art['blob'])

        # Set custom location
        url = '/sample_artifact/%s' % art['id']
        self.put(url=url + '/dict_of_blobs/blob', data=body,
                 status=200, headers=headers)

        # Get the artifact, blob property should have status 'active'
        art = self.get(url=url, status=200)
        self.assertEqual('active', art['dict_of_blobs']['blob']['status'])
        self.assertIsNotNone(art['dict_of_blobs']['blob']['md5'])
        self.assertIsNone(art['dict_of_blobs']['blob']['size'])
        self.assertIsNone(art['dict_of_blobs']['blob']['content_type'])
        self.assertEqual('https://www.apache.org/licenses/LICENSE-2.0.txt',
                         art['dict_of_blobs']['blob']['url'])
        self.assertNotIn('id', art['dict_of_blobs']['blob'])
        # test re-add failed
        self.put(url=url + '/dict_of_blobs/blob', data=body, status=409,
                 headers=headers)

        # test request failed with non-json containment
        self.put(url=url + '/dict_of_blobs/blob_incorrect', data="incorrect",
                 status=400, headers=headers)

        # delete the artifact
        self.delete(url=url)

    def test_delete_external_blob(self):
        # Create artifact
        art = self.create_artifact({'name': 'name5',
                                    'version': '1.0',
                                    'tags': ['tag1', 'tag2', 'tag3'],
                                    'int1': 2048,
                                    'float1': 987.654,
                                    'str1': 'lalala',
                                    'bool1': False,
                                    'string_required': '123'})
        self.assertIsNotNone(art['id'])

        # Set custom location
        url = '/sample_artifact/%s' % art['id']
        body = jsonutils.dumps(
            {'url': 'https://www.apache.org/licenses/LICENSE-2.0.txt',
             'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"})
        headers = {'Content-Type':
                   'application/vnd+openstack.glare-custom-location+json'}
        art = self.put(url=url + '/blob', data=body,
                       status=200, headers=headers)
        self.assertEqual('active', art['blob']['status'])
        self.assertEqual('fake', art['blob']['md5'])
        self.assertEqual('fake_sha', art['blob']['sha1'])
        self.assertEqual('fake_sha256', art['blob']['sha256'])
        self.assertIsNone(art['blob']['size'])
        self.assertIsNone(art['blob']['content_type'])
        self.assertEqual('https://www.apache.org/licenses/LICENSE-2.0.txt',
                         art['blob']['url'])
        self.assertNotIn('id', art['blob'])

        # Delete should work
        art = self.delete(url=url + '/blob', status=200)
        self.assertIsNone(art['blob'])

        # Deletion of empty blob fails
        self.delete(url=url + '/blob', status=404)

        # Deletion of non-blob field fails
        self.delete(url=url + '/int1', status=400)

        # Deletion ofn non-existing field fails
        self.delete(url=url + '/NONEXIST', status=400)

        # Upload data
        data = 'some_arbitrary_testing_data'
        headers = {'Content-Type': 'application/octet-stream'}
        art = self.put(url=url + '/blob', data=data, status=200,
                       headers=headers)
        self.assertEqual('active', art['blob']['status'])
        md5 = hashlib.md5(data.encode('UTF-8')).hexdigest()
        sha1 = hashlib.sha1(data.encode('UTF-8')).hexdigest()
        sha256 = hashlib.sha256(data.encode('UTF-8')).hexdigest()
        self.assertEqual(md5, art['blob']['md5'])
        self.assertEqual(sha1, art['blob']['sha1'])
        self.assertEqual(sha256, art['blob']['sha256'])

        # Deletion of internal blob fails
        self.delete(url=url + '/blob', status=403)

    def test_delete_external_blob_dict(self):
        # Create artifact
        art = self.create_artifact({'name': 'name5',
                                    'version': '1.0',
                                    'tags': ['tag1', 'tag2', 'tag3'],
                                    'int1': 2048,
                                    'float1': 987.654,
                                    'str1': 'lalala',
                                    'bool1': False,
                                    'string_required': '123'})
        self.assertIsNotNone(art['id'])

        # Set custom location
        url = '/sample_artifact/%s' % art['id']
        body = jsonutils.dumps(
            {'url': 'https://www.apache.org/licenses/LICENSE-2.0.txt',
             'md5': "fake", 'sha1': "fake_sha", "sha256": "fake_sha256"})
        headers = {'Content-Type':
                   'application/vnd+openstack.glare-custom-location+json'}
        art = self.put(url=url + '/dict_of_blobs/blob', data=body,
                       status=200, headers=headers)
        self.assertEqual('active', art['dict_of_blobs']['blob']['status'])
        self.assertEqual('fake', art['dict_of_blobs']['blob']['md5'])
        self.assertEqual('fake_sha', art['dict_of_blobs']['blob']['sha1'])
        self.assertEqual('fake_sha256', art['dict_of_blobs']['blob']['sha256'])
        self.assertIsNone(art['dict_of_blobs']['blob']['size'])
        self.assertIsNone(art['dict_of_blobs']['blob']['content_type'])
        self.assertEqual('https://www.apache.org/licenses/LICENSE-2.0.txt',
                         art['dict_of_blobs']['blob']['url'])
        self.assertNotIn('id', art['dict_of_blobs']['blob'])

        # Delete should work
        art = self.delete(url=url + '/dict_of_blobs/blob', status=200)
        self.assertNotIn('blob', art['dict_of_blobs'])

        # Deletion of non-existing blob fails
        self.delete(url=url + '/dict_of_blobs/NONEXIST', status=404)

        # Upload data
        data = 'some_arbitrary_testing_data'
        headers = {'Content-Type': 'application/octet-stream'}
        art = self.put(url=url + '/dict_of_blobs/blob', data=data, status=200,
                       headers=headers)
        self.assertEqual('active', art['dict_of_blobs']['blob']['status'])
        md5 = hashlib.md5(data.encode('UTF-8')).hexdigest()
        sha1 = hashlib.sha1(data.encode('UTF-8')).hexdigest()
        sha256 = hashlib.sha256(data.encode('UTF-8')).hexdigest()
        self.assertEqual(md5, art['dict_of_blobs']['blob']['md5'])
        self.assertEqual(sha1, art['dict_of_blobs']['blob']['sha1'])
        self.assertEqual(sha256, art['dict_of_blobs']['blob']['sha256'])

        # Deletion of internal blob fails
        self.delete(url=url + '/dict_of_blobs/blob', status=403)

    def test_internal_location(self):
        self.set_user('admin')
        # Create artifact
        art = self.create_artifact({'name': 'name5'})
        self.assertIsNotNone(art['id'])

        url = '/sample_artifact/%s' % art['id']
        headers = {'Content-Type':
                   'application/vnd+openstack.glare-custom-location+json'}

        # Setting locations with forbidden schemes fails
        forbidden_schemes = ('file', 'filesystem', 'swift+config', 'sql')
        for scheme in forbidden_schemes:
            body = jsonutils.dumps(
                {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
                 'location_type': 'internal',
                 'url': scheme + '://FAKE_LOCATION.com'})
            self.put(url=url + '/blob', data=body, status=403, headers=headers)

        # Setting locations with unknown schemes fail
        body = jsonutils.dumps(
            {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
             'location_type': 'internal',
             'url': 'UNKNOWN://FAKE_LOCATION.com'})
        self.put(url=url + '/blob', data=body, status=400, headers=headers)

        body = jsonutils.dumps(
            {'md5': 'fake', 'sha1': 'fake_sha', 'sha256': 'fake_sha256',
             'location_type': 'internal',
             'url': 'https://FAKE_LOCATION.com'})
        art = self.put(url=url + '/blob', data=body, status=200,
                       headers=headers)

        self.assertFalse(art['blob']['external'])
        self.assertEqual('active', art['blob']['status'])
        self.assertEqual('fake', art['blob']['md5'])
        self.assertEqual('fake_sha', art['blob']['sha1'])
        self.assertEqual('fake_sha256', art['blob']['sha256'])
        self.assertIsNone(art['blob']['size'])
        self.assertIsNone(art['blob']['content_type'])
        self.assertEqual('/artifacts/sample_artifact/%s/blob' % art['id'],
                         art['blob']['url'])
        self.assertNotIn('id', art['blob'])


class TestTags(base.TestArtifact):
    def test_tags(self):
        # Create artifact
        art = self.create_artifact({'name': 'name5',
                                    'version': '1.0',
                                    'tags': ['tag1', 'tag2', 'tag3'],
                                    'int1': 2048,
                                    'float1': 987.654,
                                    'str1': 'lalala',
                                    'bool1': False,
                                    'string_required': '123'})
        self.assertIsNotNone(art['id'])

        url = '/sample_artifact/%s' % art['id']
        data = [{
            "op": "replace",
            "path": "/status",
            "value": "active"
        }]
        art = self.patch(url=url, data=data, status=200)
        self.assertEqual('active', art['status'])
        art = self.admin_action(art['id'], self.make_public)

        self.assertEqual('public', art['visibility'])
        # only admins can update tags for public artifacts
        self.set_user("admin")

        # Check that tags created correctly
        url = '/sample_artifact/%s' % art['id']
        resp = self.get(url=url, status=200)
        for tag in ['tag1', 'tag2', 'tag3']:
            self.assertIn(tag, resp['tags'])

        # Set new tag list to the art
        body = [{"op": "replace",
                 "path": "/tags",
                 "value": ["new_tag1", "new_tag2", "new_tag3"]}]
        resp = self.patch(url=url, data=body, status=200)
        for tag in ['new_tag1', 'new_tag2', 'new_tag3']:
            self.assertIn(tag, resp['tags'])

        # Delete all tags from the art
        body = [{"op": "replace",
                 "path": "/tags",
                 "value": []}]
        resp = self.patch(url=url, data=body, status=200)
        self.assertEqual([], resp['tags'])

        # Set new tags as null
        body = [{"op": "replace",
                 "path": "/tags",
                 "value": None}]
        resp = self.patch(url=url, data=body, status=200)
        self.assertEqual([], resp['tags'])

        # Get the list of tags
        resp = self.get(url=url, status=200)
        self.assertEqual([], resp['tags'])


class TestArtifactOps(base.TestArtifact):
    def test_create(self):
        """All tests related to artifact creation"""
        # check that cannot create artifact for non-existent artifact type
        self.post('/incorrect_artifact', {"name": "t"}, status=404)
        # check that cannot accept non-json body
        self.post('/incorrect_artifact', "incorrect_body", status=400)
        # check that cannot accept incorrect content type
        self.post('/sample_artifact', {"name": "t"}, status=415,
                  headers={"Content-Type": "application/octet-stream"})
        # check that cannot create artifact without name
        self.create_artifact(data={"int1": 1024}, status=400)
        # check that cannot create artifact with too long name
        self.create_artifact(data={"name": "t" * 256}, status=400)
        # check that cannot create artifact with empty name
        self.create_artifact(data={"name": ""}, status=400)
        # check that can create af without version
        private_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str"})
        # check that default is set on artifact create
        uuid.UUID(private_art['id'])
        self.assertEqual('0.0.0', private_art['version'])
        self.assertEqual("default", private_art["system_attribute"])
        self.assertEqual(self.users['user1']['tenant_id'],
                         private_art['owner'])

        # check that cannot create artifact with invalid version
        self.create_artifact(data={"name": "test_af",
                                   "version": "dummy_version"}, status=400)
        # check that cannot create artifact with empty and long version
        self.create_artifact(data={"name": "test_af",
                                   "version": ""}, status=400)
        # check that cannot create artifact with empty and long version
        self.create_artifact(data={"name": "test_af",
                                   "version": "t" * 256}, status=400)
        # check that artifact artifact with the same name-version cannot
        # be created
        self.create_artifact(data={"name": "test_af"}, status=409)
        # check that we cannot create af with the same version but different
        # presentation
        self.create_artifact(data={"name": "test_af", "version": "0.0"},
                             status=409)
        # check that we can create artifact with different version and tags
        new_af = self.create_artifact(
            data={"name": "test_af", "version": "0.0.1",
                  "tags": ["tag1", "tag2"]})
        self.assertEqual({"tag1", "tag2"}, set(new_af["tags"]))
        # check that we cannot create artifact with visibility
        self.create_artifact(data={"name": "test_af", "version": "0.0.2",
                                   "visibility": "private"}, status=400)
        # check that we cannot create artifact with system property
        self.create_artifact(data={"name": "test_af", "version": "0.0.2",
                                   "system_attribute": "test"}, status=403)
        # check that we cannot specify blob in create
        self.create_artifact(data={"name": "test_af", "version": "0.0.2",
                                   "blob": {
                                       'url': None, 'size': None,
                                       'md5': None, 'status': 'saving',
                                       'external': False}}, status=400)
        # check that anonymous user cannot create artifact
        self.set_user("anonymous")
        self.create_artifact(data={"name": "test_af", "version": "0.0.2"},
                             status=403)
        # check that another user can create artifact
        # with the same name version
        self.set_user("user2")
        some_af = self.create_artifact(data={"name": "test_af"})

        # check we can create artifact with all available attributes
        # (except blobs and system)
        expected = {
            "name": "test_big_create",
            "link1": "/artifacts/sample_artifact/%s" % some_af['id'],
            "bool1": True,
            "int1": 2323,
            "float1": 0.1,
            "str1": "test",
            "list_of_str": ["test"],
            "list_of_int": [0],
            "dict_of_str": {"test": "test"},
            "dict_of_int": {"test": 0},
            "string_mutable": "test",
            "string_required": "test",
        }
        big_af = self.create_artifact(data=expected)
        actual = {}
        for k in expected:
            actual[k] = big_af[k]
        self.assertEqual(expected, actual)
        # check that we cannot access artifact from other user
        # check that active artifact is not available for other user
        url = '/sample_artifact/%s' % private_art['id']
        self.get(url, status=404)
        # check we cannot create af with non-existing property
        self.create_artifact(data={"name": "test_af_ne",
                                   "non_exist": "non_exist"}, status=400)
        # activate and publish artifact to check that we can create
        # private artifact with the same name version
        self.set_user("user1")

        self.patch(url=url, data=self.make_active)
        self.admin_action(private_art['id'], self.make_public)
        self.create_artifact(data={"name": "test_af",
                                   "string_required": "test_str"})

    def test_activate(self):
        # create artifact to update
        private_art = self.create_artifact(
            data={"name": "test_af",
                  "version": "0.0.1"})
        # cannot activate artifact without required for activate attributes
        url = '/sample_artifact/%s' % private_art['id']
        self.patch(url=url, data=self.make_active, status=403)
        add_required = [{
            "op": "replace",
            "path": "/string_required",
            "value": "string"
        }]
        self.patch(url=url, data=add_required)
        # can activate if body contains non status changes
        make_active_with_updates = self.make_active + [{"op": "replace",
                                                        "path": "/description",
                                                        "value": "test"}]
        active_art = self.patch(url=url, data=make_active_with_updates)
        private_art['status'] = 'active'
        private_art['activated_at'] = active_art['activated_at']
        private_art['updated_at'] = active_art['updated_at']
        private_art['string_required'] = 'string'
        private_art['description'] = 'test'
        self.assertEqual(private_art, active_art)
        # check that active artifact is not available for other user
        self.set_user("user2")
        self.get(url, status=404)
        self.set_user("user1")

        # test that activate is idempotent
        self.patch(url=url, data=self.make_active)
        # test activate deleted artifact
        self.delete(url=url)
        self.patch(url=url, data=self.make_active, status=404)

    def test_publish(self):
        # create artifact to update
        self.set_user('admin')
        private_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})

        url = '/sample_artifact/%s' % private_art['id']
        # test that we cannot publish drafted artifact
        self.patch(url=url, data=self.make_public, status=403)

        self.patch(url=url, data=self.make_active)

        # test that cannot publish deactivated artifact
        self.patch(url, data=self.make_deactivated)
        self.patch(url, data=self.make_public, status=403)

        self.patch(url=url, data=self.make_active)

        # test that visibility can be specified in the request with
        # other updates
        make_public_with_updates = self.make_public + [
            {"op": "replace",
             "path": "/string_mutable",
             "value": "test"}]
        self.patch(url=url, data=make_public_with_updates)
        # check public artifact
        public_art = self.patch(url=url, data=self.make_public)
        private_art['activated_at'] = public_art['activated_at']
        private_art['visibility'] = 'public'
        private_art['status'] = 'active'
        private_art['updated_at'] = public_art['updated_at']
        private_art['string_mutable'] = 'test'
        self.assertEqual(private_art, public_art)
        # check that public artifact available for simple user
        self.set_user("user1")
        self.get(url)
        self.set_user("admin")
        # test that artifact publish with the same name and version failed
        duplicate_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        dup_url = '/sample_artifact/%s' % duplicate_art['id']
        # proceed with duplicate testing
        self.patch(url=dup_url, data=self.make_active)
        self.patch(url=dup_url, data=self.make_public, status=409)

    def test_delete(self):
        # try ro delete not existing artifact
        url = '/sample_artifact/111111'
        self.delete(url=url, status=404)

        # check that we can delete artifact with soft link
        art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        artd = self.create_artifact(
            data={"name": "test_afd", "string_required": "test_str",
                  "version": "0.0.1",
                  "link1": '/artifacts/sample_artifact/%s' % art['id']})

        url = '/sample_artifact/%s' % artd['id']
        self.delete(url=url, status=204)

        # try to change status of artifact to deleting
        url = '/sample_artifact/%s' % art['id']
        patch = [{'op': 'replace',
                  'value': 'deleting',
                  'path': '/status'}]
        self.patch(url=url, data=patch, status=400)

        # delete artifact via different user (non admin)
        self.set_user('user2')
        self.delete(url=url, status=404)

        # delete artifact via admin user
        self.set_user('admin')
        self.delete(url=url, status=204)

        # delete public artifact via different user
        self.set_user('user1')
        art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        url = '/sample_artifact/%s' % art['id']
        self.patch(url=url, data=self.make_active)
        self.admin_action(art['id'], self.make_public)
        self.set_user('user2')
        self.delete(url=url, status=403)

        self.set_user('user1')
        self.delete(url=url, status=403)
        self.set_user('admin')
        self.delete(url=url)

        # delete deactivated artifact
        art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        url = '/sample_artifact/%s' % art['id']
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_deactivated)
        self.delete(url=url, status=204)
        self.get(url=url, status=404)
        self.assertEqual(0, len(self.get(
            url='/sample_artifact')['sample_artifact']))

    def test_deactivate(self):
        # test artifact deactivate for non-active artifact
        private_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        url = '/sample_artifact/%s' % private_art['id']
        self.admin_action(private_art['id'], self.make_deactivated, 403)
        self.patch(url, self.make_active)
        self.set_user('admin')
        # test can deactivate if there is something else in request
        make_deactived_with_updates = [
            {"op": "replace",
             "path": "/description",
             "value": "test"}] + self.make_deactivated
        # test artifact deactivate success
        deactivated_art = self.admin_action(
            private_art['id'], make_deactived_with_updates)
        self.assertEqual("deactivated", deactivated_art["status"])
        self.assertEqual("test", deactivated_art["description"])
        # test deactivate is idempotent
        self.patch(url, self.make_deactivated)

    def test_reactivate(self):
        self.set_user('admin')
        private_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})
        url = '/sample_artifact/%s' % private_art['id']
        self.patch(url, self.make_active)
        self.admin_action(private_art['id'], self.make_deactivated)
        # test can reactivate if there is something else in request
        make_reactived_with_updates = self.make_active + [
            {"op": "replace",
             "path": "/description",
             "value": "test"}]
        # test artifact deactivate success
        reactivated_art = self.admin_action(
            private_art['id'], make_reactived_with_updates)
        self.assertEqual("active", reactivated_art["status"])
        self.assertEqual("test", reactivated_art["description"])


class TestUpdate(base.TestArtifact):
    def test_update_artifact_before_activate(self):
        """Test updates for artifact before activation"""
        # create artifact to update
        private_art = self.create_artifact(data={"name": "test_af"})
        url = '/sample_artifact/%s' % private_art['id']
        # check we can update artifact
        change_version = [{
            "op": "replace",
            "path": "/version",
            "value": "0.0.2"
        }]
        self.patch(url=url, data=change_version)

        # wrong patch format fails with 400 error
        invalid_patch = {
            "op": "replace",
            "path": "/version",
            "value": "0.0.2"
        }
        self.patch(url=url, data=invalid_patch, status=400)

        # check that we cannot update af if af with
        # the same name or version exists
        dup_version = self.create_artifact(
            data={"name": "test_af", "version": "0.0.1"})
        dupv_url = '/sample_artifact/%s' % dup_version['id']
        change_version_dup = [{
            "op": "replace",
            "path": "/version",
            "value": "0.0.2"
        }]
        self.patch(url=dupv_url, data=change_version_dup, status=409)

        dup_name = self.create_artifact(data={"name": "test_name_af",
                                              "version": "0.0.2"})
        dupn_url = '/sample_artifact/%s' % dup_name['id']
        change_name = [{
            "op": "replace",
            "path": "/name",
            "value": "test_af"
        }]
        self.patch(url=dupn_url, data=change_name, status=409)
        # check that we can update artifacts dup
        # after first artifact updated name and version
        change_version[0]['value'] = "0.0.3"
        self.patch(url=url, data=change_version)
        self.patch(url=dupn_url, data=change_name)
        # check that we can update artifact dupv to target version
        # also check that after deletion of artifact with the same name
        # version I can update dupv
        self.delete(dupn_url)
        self.patch(url=dupv_url, data=change_version_dup)
        # check we cannot update artifact with incorrect content-type
        self.patch(url, {}, status=415,
                   headers={"Content-Type": "application/json"})
        # check we cannot update tags with patch
        set_tags = [{
            "op": "replace",
            "path": "/tags",
            "value": "test_af"
        }]
        self.patch(url, set_tags, status=400)
        # check we cannot update artifact with incorrect json-patch
        self.patch(url, "incorrect json patch", status=400)
        # check update is correct if there is no update
        no_name_update = [{
            "op": "replace",
            "path": "/name",
            "value": "test_af"
        }]
        self.patch(url, no_name_update)
        # check add new property request rejected
        add_prop = [{
            "op": "add",
            "path": "/string1",
            "value": "test_af"
        }]
        self.patch(url, add_prop, 400)
        # check delete property request rejected
        add_prop[0]["op"] = "remove"
        add_prop[0]["path"] = "/string_required"
        self.patch(url, add_prop, 400)
        # check we cannot update system attr with patch
        system_attr = [{
            "op": "replace",
            "path": "/system_attribute",
            "value": "dummy"
        }]
        self.patch(url, system_attr, 403)
        # check cannot update blob attr with patch
        blob_attr = [{
            "op": "replace",
            "path": "/blob",
            "value": {"name": "test_af", "version": "0.0.2",
                      "blob": {'url': None, 'size': None, 'md5': None,
                               'status': 'saving', 'external': False}}}]
        self.patch(url, blob_attr, 400)
        blob_attr[0]["path"] = "/dict_of_blobs/-"
        blob_attr[0]["op"] = "add"
        self.patch(url, blob_attr, 400)
        # test update correctness for all attributes
        big_update_patch = [
            {"op": "replace", "path": "/bool1", "value": True},
            {"op": "replace", "path": "/int1", "value": 2323},
            {"op": "replace", "path": "/float1", "value": 0.1},
            {"op": "replace", "path": "/str1", "value": "test"},
            {"op": "replace", "path": "/list_of_str", "value": ["test"]},
            {"op": "replace", "path": "/list_of_int", "value": [0]},
            {"op": "replace", "path": "/dict_of_str",
             "value": {"test": "test"}},
            {"op": "replace", "path": "/dict_of_int",
             "value": {"test": 0}},
            {"op": "replace", "path": "/string_mutable", "value": "test"},
            {"op": "replace", "path": "/string_required", "value": "test"},
        ]
        upd_af = self.patch(url, big_update_patch)
        for patch_item in big_update_patch:
            self.assertEqual(patch_item.get("value"),
                             upd_af[patch_item.get("path")[1:]])

        # check we can update private artifact
        # to the same name version as public artifact
        self.patch(url=url, data=self.make_active)
        self.admin_action(private_art['id'], self.make_public)
        self.patch(url=dupv_url, data=change_version)

    def test_update_after_activate_and_publish(self):
        # activate artifact
        private_art = self.create_artifact(
            data={"name": "test_af", "string_required": "test_str",
                  "version": "0.0.1"})

        url = '/sample_artifact/%s' % private_art['id']
        self.patch(url=url, data=self.make_active)
        # test that immutable properties cannot be updated
        upd_immutable = [{
            "op": "replace",
            "path": "/name",
            "value": "new_name"
        }]
        self.patch(url, upd_immutable, status=403)
        # test that mutable properties can be updated
        upd_mutable = [{
            "op": "replace",
            "path": "/string_mutable",
            "value": "new_value"
        }]
        updated_af = self.patch(url, upd_mutable)
        self.assertEqual("new_value", updated_af["string_mutable"])
        # test cannot update deactivated artifact
        upd_mutable[0]["value"] = "another_new_value"
        self.admin_action(private_art['id'], self.make_deactivated)
        # test that nobody(even admin) can publish deactivated artifact
        self.set_user("admin")
        self.patch(url, self.make_public, 403)
        self.set_user("user1")
        self.patch(url, upd_mutable, 403)
        self.admin_action(private_art['id'], self.make_active)
        # publish artifact
        self.admin_action(private_art['id'], self.make_public)
        # check we cannot update public artifact anymore
        self.patch(url, upd_mutable, status=403)
        self.patch(url, upd_mutable, status=403)
        # check that admin can update public artifact
        self.set_user("admin")
        self.patch(url, upd_mutable)

    def test_update_with_validators(self):
        data = {'name': 'test_af',
                'version': '0.0.1',
                'list_validators': ['a', 'b', 'c'],
                'dict_validators': {'abc': 'a', 'def': 'b'}}
        art = self.create_artifact(data=data)
        url = '/sample_artifact/%s' % art['id']

        # min int_validators value is 10
        patch = [{"op": "replace", "path": "/int_validators", "value": 9}]
        self.patch(url=url, data=patch, status=400)

        # max int_validators value is 20
        patch = [{"op": "replace", "path": "/int_validators", "value": 21}]
        self.patch(url=url, data=patch, status=400)

        # number 15 is okay
        patch = [{"op": "replace", "path": "/int_validators", "value": 15}]
        self.patch(url=url, data=patch, status=200)

        # max string length is 255
        patch = [{"op": "replace", "path": "/str1", "value": 'd' * 256}]
        self.patch(url=url, data=patch, status=400)

        # 'cc' is not allowed value for the string
        patch = [{"op": "replace", "path": "/string_validators",
                  "value": 'cc'}]
        self.patch(url=url, data=patch, status=400)

        # 'aa' is okay
        patch = [{"op": "replace", "path": "/string_validators",
                  "value": 'aa'}]
        self.patch(url=url, data=patch)

        # 'bb' is okay too
        patch = [{"op": "replace", "path": "/string_validators",
                  "value": 'bb'}]
        self.patch(url=url, data=patch)

        # even if 'c' * 11 is allowed value it exceeds MaxLen's 10 character
        # limit
        patch = [{"op": "replace", "path": "/string_validators",
                  "value": 'c' * 11}]
        self.patch(url=url, data=patch, status=400)

        # string_regex format it '^([0-9a-fA-F]){8}$'
        patch = [{"op": "replace", "path": "/string_regex",
                  "value": 'INVALID'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/string_regex",
                  "value": '167f808Z'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/string_regex",
                  "value": '167f80835'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/string_regex",
                  "value": '167f8083'}]
        self.patch(url=url, data=patch)

        # test list has 3 elements maximum
        patch = [{"op": "add", "path": "/list_validators/-", "value": 'd'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/list_validators",
                  "value": ['a', 'b', 'c', 'd']}]
        self.patch(url=url, data=patch, status=400)

        # test list values are unique
        patch = [{"op": "replace", "path": "/list_validators/2", "value": 'b'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/list_validators",
                  "value": ['a', 'b', 'b']}]
        self.patch(url=url, data=patch, status=400)

        # regular update works
        patch = [{"op": "replace", "path": "/list_validators/1", "value": 'd'}]
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['list_validators'], ['a', 'd', 'c'])

        patch = [{"op": "replace", "path": "/list_validators",
                  "value": ['c', 'b', 'a']}]
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['list_validators'], ['c', 'b', 'a'])

        # test adding wrong key to dict
        patch = [{"op": "add", "path": "/dict_validators/aaa", "value": 'b'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/dict_validators",
                  "value": {'abc': 'a', 'def': 'b', 'aaa': 'c'}}]
        self.patch(url=url, data=patch, status=400)

        # test dict has 3 elements maximum
        patch = [{"op": "add", "path": "/dict_validators/ghi", "value": 'd'}]
        self.patch(url=url, data=patch)

        patch = [{"op": "add", "path": "/dict_validators/jkl", "value": 'd'}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace", "path": "/dict_validators",
                  "value": {'abc': 'a', 'def': 'b', 'ghi': 'c', 'jkl': 'd'}}]
        self.patch(url=url, data=patch, status=400)

        # regular update works
        patch = [{"op": "replace", "path": "/dict_validators/abc",
                  "value": "q"}]
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['dict_validators'],
                         {'abc': 'q', 'def': 'b', 'ghi': 'd'})

        patch = [{"op": "replace", "path": "/dict_validators",
                  "value": {'abc': 'l', 'def': 'x', 'ghi': 'z'}}]
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['dict_validators'],
                         {'abc': 'l', 'def': 'x', 'ghi': 'z'})

    def test_update_base_fields(self):
        data = {'name': 'test_af',
                'version': '0.0.1'}
        art = self.create_artifact(data=data)
        url = '/sample_artifact/%s' % art['id']

        # INT
        # float to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": 1.1}]
        art = self.patch(url=url, data=patch)
        self.assertEqual(1, art['int1'])

        # str(int) to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": '2'}]
        art = self.patch(url=url, data=patch)
        self.assertEqual(2, art['int1'])

        # str(float) to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": '3.0'}]
        self.patch(url=url, data=patch, status=400)

        # str(int) to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": ''}]
        self.patch(url=url, data=patch, status=400)

        # empty list to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": []}]
        self.patch(url=url, data=patch, status=400)

        # empty dict to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": {}}]
        self.patch(url=url, data=patch, status=400)

        # bool to int
        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": True}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(1, art['int1'])

        patch = [{"op": "replace",
                  "path": "/int1",
                  "value": False}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(0, art['int1'])

        # FLOAT
        # int to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": 1}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(1.0, art['float1'])

        # str(int) to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": '2'}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(2.0, art['float1'])

        # str(int) to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": []}]
        self.patch(url=url, data=patch, status=400)

        # str(int) to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": {}}]
        self.patch(url=url, data=patch, status=400)

        # str(bool) to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": 'True'}]
        self.patch(url=url, data=patch, status=400)

        # bool to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": True}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(1.0, art['float1'])

        # str(float) to float
        patch = [{"op": "replace",
                  "path": "/float1",
                  "value": '3.0'}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(3.0, art['float1'])

        # STRING
        # str to str
        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": '3.0'}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual('3.0', art['str1'])

        # int to str
        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": 1}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual('1', art['str1'])

        # float to str
        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": 1.0}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual('1.0', art['str1'])

        # bool to str
        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": True}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual('True', art['str1'])

        # empty list to str
        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": []}]
        self.patch(url=url, data=patch, status=400)

        patch = [{"op": "replace",
                  "path": "/str1",
                  "value": {}}]
        self.patch(url=url, data=patch, status=400)

        # BOOL
        # int to bool
        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": 1}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(True, art['bool1'])

        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": 0}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        # float to bool
        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": 2.1}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": 1.1}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        # string to bool
        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": '1'}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(True, art['bool1'])

        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": ''}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])
        # [] to bool
        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": []}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": [1]}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])
        # {} to bool
        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": {}}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        patch = [{"op": "replace",
                  "path": "/bool1",
                  "value": {'1', 1}}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(False, art['bool1'])

        # LIST OF STR AND INT
        # {} to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": {}}]
        self.patch(url=url, data=patch, status=400)

        # [] to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": []}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual([], art['list_of_str'])

        # list of int to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": [1, 2, 3]}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(['1', '2', '3'], art['list_of_str'])

        # list of bool to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": [True, False, True]}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual(['True', 'False', 'True'], art['list_of_str'])

        # str to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": '123'}]
        self.patch(url=url, data=patch, status=400)

        # int to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": 11}]
        self.patch(url=url, data=patch, status=400)

        # bool to list of str
        patch = [{"op": "replace",
                  "path": "/list_of_str",
                  "value": True}]
        self.patch(url=url, data=patch, status=400)

        # Dict OF INT
        # [] to dict of int
        patch = [{"op": "replace",
                  "path": "/dict_of_int",
                  "value": []}]
        self.patch(url=url, data=patch, status=400)

        # {} to dict of int
        patch = [{"op": "replace",
                  "path": "/dict_of_int",
                  "value": {}}]
        art = self.patch(url=url, data=patch, status=200)
        self.assertEqual({}, art['dict_of_int'])

        # int to dict of int
        patch = [{"op": "replace",
                  "path": "/dict_of_int",
                  "value": 1}]
        self.patch(url=url, data=patch, status=400)

        # bool to dict of int
        patch = [{"op": "replace",
                  "path": "/dict_of_int",
                  "value": True}]
        self.patch(url=url, data=patch, status=400)

        # string to dict of int
        patch = [{"op": "replace",
                  "path": "/dict_of_int",
                  "value": 'aaa'}]
        self.patch(url=url, data=patch, status=400)

    def test_update_field_dict(self):
        art1 = self.create_artifact(data={"name": "art1"})

        # create artifact without dict prop
        data = {'name': 'art_without_dict'}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual({}, result['dict_of_str'])

        # create artifact with dict prop
        data = {'name': 'art_with_dict',
                'dict_of_str': {'a': '1', 'b': '3'}}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual({'a': '1', 'b': '3'}, result['dict_of_str'])

        # create artifact with empty dict
        data = {'name': 'art_with_empty_dict',
                'dict_of_str': {}}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual({}, result['dict_of_str'])

        # add element in invalid path
        data = [{'op': 'add',
                 'path': '/dict_of_str',
                 'value': 'val1'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # add new element
        data = [{'op': 'add',
                 'path': '/dict_of_str/new',
                 'value': 'val1'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual('val1', result['dict_of_str']['new'])

        # add existent element
        data = [{'op': 'add',
                 'path': '/dict_of_str/new',
                 'value': 'val_new'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual('val_new', result['dict_of_str']['new'])

        # add element with empty key
        data = [{'op': 'add',
                 'path': '/dict_of_str/',
                 'value': 'val1'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # replace element
        data = [{'op': 'replace',
                 'path': '/dict_of_str/new',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual('val2', result['dict_of_str']['new'])

        # replace non-existent element
        data = [{'op': 'replace',
                 'path': '/dict_of_str/non_exist',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # remove element
        data = [{'op': 'remove',
                 'path': '/dict_of_str/new',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertIsNone(result['dict_of_str'].get('new'))

        # remove non-existent element
        data = [{'op': 'remove',
                 'path': '/dict_of_str/non_exist',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # set value
        data = [{'op': 'add',
                 'path': '/dict_of_str',
                 'value': {'key1': 'val1', 'key2': 'val2'}}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual({'key1': 'val1', 'key2': 'val2'},
                         result['dict_of_str'])

        # replace value
        data = [{'op': 'add',
                 'path': '/dict_of_str',
                 'value': {'key11': 'val1', 'key22': 'val2'}}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual({'key11': 'val1', 'key22': 'val2'},
                         result['dict_of_str'])

        # remove value
        data = [{'op': 'add',
                 'path': '/dict_of_str',
                 'value': {}}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual({},
                         result['dict_of_str'])

        # set an element of the wrong non-conversion type value
        data = [{'op': 'add',
                 'path': '/dict_of_str/wrong_type',
                 'value': [1, 2, 4]}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # set an element of the wrong conversion type value
        data = [{'op': 'add',
                 'path': '/dict_of_str/wrong_type',
                 'value': 1}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual('1', result['dict_of_str']['wrong_type'])

        # add element with None value
        data = [{'op': 'add',
                 'path': '/dict_of_blob/nane_value',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

    def test_update_field_list(self):
        art1 = self.create_artifact(data={"name": "art1"})

        # create artifact without list prop
        data = {'name': 'art_without_list'}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual([], result['list_of_str'])

        # create artifact with list prop
        data = {'name': 'art_with_list',
                'list_of_str': ['a', 'b']}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual(['a', 'b'], result['list_of_str'])

        # create artifact with empty list
        data = {'name': 'art_with_empty_list',
                'list_of_str': []}
        result = self.post(url='/sample_artifact', status=201, data=data)
        self.assertEqual([], result['list_of_str'])

        # add value
        data = [{'op': 'add',
                 'path': '/list_of_str',
                 'value': ['b', 'd']}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['b', 'd'], result['list_of_str'])

        # replace value
        data = [{'op': 'replace',
                 'path': '/list_of_str',
                 'value': ['aa', 'dd']}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['aa', 'dd'], result['list_of_str'])

        # remove value
        data = [{'op': 'add',
                 'path': '/list_of_str',
                 'value': []}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual([], result['list_of_str'])

        # add new element on empty list
        self.assertEqual([], art1['list_of_str'])
        data = [{'op': 'add',
                 'path': '/list_of_str/-',
                 'value': 'val1'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val1'], result['list_of_str'])

        # add new element on index
        data = [{'op': 'add',
                 'path': '/list_of_str/0',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val2', 'val1'], result['list_of_str'])

        # add new element on next index
        data = [{'op': 'add',
                 'path': '/list_of_str/1',
                 'value': 'val3'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val2', 'val3', 'val1'], result['list_of_str'])

        # add new element on default index
        data = [{'op': 'add',
                 'path': '/list_of_str/-',
                 'value': 'val4'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val2', 'val3', 'val1', 'val4'],
                         result['list_of_str'])

        # add new element on non-existent index
        data = [{'op': 'add',
                 'path': '/list_of_str/10',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # replace element on index
        data = [{'op': 'replace',
                 'path': '/list_of_str/1',
                 'value': 'val_new'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val2', 'val_new', 'val1', 'val4'],
                         result['list_of_str'])

        # replace element on default index
        data = [{'op': 'replace',
                 'path': '/list_of_str/-',
                 'value': 'val-'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # replace new element on non-existent index
        data = [{'op': 'replace',
                 'path': '/list_of_str/99',
                 'value': 'val_new'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # remove element on index
        data = [{'op': 'remove',
                 'path': '/list_of_str/1',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(['val2', 'val1', 'val4'], result['list_of_str'])

        # remove element on default index
        data = [{'op': 'remove',
                 'path': '/list_of_str/-',
                 'value': 'val3'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        # remove new element on non-existent index
        data = [{'op': 'remove',
                 'path': '/list_of_str/999',
                 'value': 'val2'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

    def test_update_remove_properties(self):
        data = {
            "name": "test_big_create",
            "version": "1.0.0",
            "bool1": True,
            "int1": 2323,
            "float1": 0.1,
            "str1": "test",
            "list_of_str": ["test1", "test2"],
            "list_of_int": [0, 1, 2],
            "dict_of_str": {"test": "test"},
            "dict_of_int": {"test": 0},
            "string_mutable": "test",
            "string_required": "test",
        }
        art1 = self.create_artifact(data=data)

        # remove the whole list of strings
        data = [{'op': 'replace',
                 'path': '/list_of_str',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual([], result['list_of_str'])

        # remove the whole list of ints
        data = [{'op': 'replace',
                 'path': '/list_of_int',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual([], result['list_of_int'])

        # remove the whole dict of strings
        data = [{'op': 'replace',
                 'path': '/dict_of_str',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual({}, result['dict_of_str'])

        # remove the whole dict of ints
        data = [{'op': 'replace',
                 'path': '/dict_of_int',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual({}, result['dict_of_int'])

        # remove bool1
        data = [{'op': 'replace',
                 'path': '/bool1',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertEqual(False, result['bool1'])

        # remove int1
        data = [{'op': 'replace',
                 'path': '/int1',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertIsNone(result['int1'])

        # remove float1
        data = [{'op': 'replace',
                 'path': '/float1',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        result = self.patch(url=url, data=data)
        self.assertIsNone(result['float1'])

        # cannot remove id, because it's a system field
        data = [{'op': 'replace',
                 'path': '/id',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=403)

        # cannot remove name
        data = [{'op': 'replace',
                 'path': '/name',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        headers = {'Content-Type': 'application/octet-stream'}
        self.put(url=url + '/blob', data="d" * 1000, headers=headers)

        # cannot remove blob
        data = [{'op': 'replace',
                 'path': '/blob',
                 'value': None}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

    def test_update_malformed_json_patch(self):
        data = {'name': 'ttt'}
        art1 = self.create_artifact(data=data)

        data = [{'op': 'replace', 'path': None, 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '/', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '//', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': 'name/', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '*/*', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': None, 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': '/', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': '//', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': 'name/', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': '*/*', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'add', 'path': '/name'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': None}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '/'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '//'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': 'name/'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'replace', 'path': '*/*'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)

        data = [{'op': 'no-op', 'path': '/name', 'value': 'aaa'}]
        url = '/sample_artifact/%s' % art1['id']
        self.patch(url=url, data=data, status=400)


class TestLinks(base.TestArtifact):
    def test_manage_links(self):
        some_af = self.create_artifact(data={"name": "test_af"})
        dep_af = self.create_artifact(data={"name": "test_dep_af"})
        dep_url = "/artifacts/sample_artifact/%s" % some_af['id']

        # set valid link
        patch = [{"op": "replace", "path": "/link1", "value": dep_url}]
        url = '/sample_artifact/%s' % dep_af['id']
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['link1'], dep_url)

        # remove link from artifact
        patch = [{"op": "replace", "path": "/link1", "value": None}]
        af = self.patch(url=url, data=patch)
        self.assertIsNone(af['link1'])

        # try to set invalid link
        patch = [{"op": "replace", "path": "/link1", "value": "Invalid"}]
        self.patch(url=url, data=patch, status=400)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid.uuid4()
        patch = [{"op": "replace",
                  "path": "/link1",
                  "value": non_exiting_url}]
        self.patch(url=url, data=patch, status=400)

    def test_manage_dict_of_links(self):
        some_af = self.create_artifact(data={"name": "test_af"})
        dep_af = self.create_artifact(data={"name": "test_dep_af"})
        dep_url = "/artifacts/sample_artifact/%s" % some_af['id']

        # set valid link
        patch = [{"op": "add",
                  "path": "/dict_of_links/link1",
                  "value": dep_url}]
        url = '/sample_artifact/%s' % dep_af['id']
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['dict_of_links']['link1'], dep_url)

        # remove link from artifact
        patch = [{"op": "remove",
                  "path": "/dict_of_links/link1"}]
        af = self.patch(url=url, data=patch)
        self.assertNotIn('link1', af['dict_of_links'])

        # try to set invalid link
        patch = [{"op": "replace",
                  "path": "/dict_of_links/link1",
                  "value": "Invalid"}]
        self.patch(url=url, data=patch, status=400)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid.uuid4()
        patch = [{"op": "replace",
                  "path": "/dict_of_links/link1",
                  "value": non_exiting_url}]
        self.patch(url=url, data=patch, status=400)

    def test_manage_list_of_links(self):
        some_af = self.create_artifact(data={"name": "test_af"})
        dep_af = self.create_artifact(data={"name": "test_dep_af"})
        dep_url = "/artifacts/sample_artifact/%s" % some_af['id']

        # set valid link
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": dep_url}]
        url = '/sample_artifact/%s' % dep_af['id']
        af = self.patch(url=url, data=patch)
        self.assertEqual(af['list_of_links'][0], dep_url)

        # remove link from artifact
        patch = [{"op": "remove",
                  "path": "/list_of_links/0"}]
        af = self.patch(url=url, data=patch)
        self.assertEqual(0, len(af['list_of_links']))

        # try to set invalid link
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": "Invalid"}]
        self.patch(url=url, data=patch, status=400)

        # try to set link to non-existing artifact
        non_exiting_url = "/artifacts/sample_artifact/%s" % uuid.uuid4()
        patch = [{"op": "add",
                  "path": "/list_of_links/-",
                  "value": non_exiting_url}]
        self.patch(url=url, data=patch, status=400)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import fixtures
import glance_store as store
from glance_store import location
import jsonpatch
from oslo_config import cfg
from oslo_config import fixture as cfg_fixture
from oslo_policy import policy as os_policy
from oslo_utils import uuidutils
import testtools

from glare.api.middleware import context
from glare.api.v1 import resource
from glare.common import policy
from glare.common import wsgi
from glare.db.sqlalchemy import api as db_api

CONF = cfg.CONF


class BaseTestCase(testtools.TestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self._config_fixture = self.useFixture(cfg_fixture.Config())

        self.users = {
            'user1': {
                'id': uuidutils.generate_uuid(),
                'tenant_id': uuidutils.generate_uuid(),
                'token': uuidutils.generate_uuid(),
                'roles': ['member']
            },
            'user2': {
                'id': uuidutils.generate_uuid(),
                'tenant_id': uuidutils.generate_uuid(),
                'token': uuidutils.generate_uuid(),
                'roles': ['member']
            },
            'admin': {
                'id': uuidutils.generate_uuid(),
                'tenant_id': uuidutils.generate_uuid(),
                'token': uuidutils.generate_uuid(),
                'roles': ['admin']
            },
            'anonymous': {
                'id': None,
                'tenant_id': None,
                'token': None,
                'roles': []
            }
        }

        self.test_dir = self.useFixture(fixtures.TempDir()).path

        CONF.set_default('connection', 'sqlite://', group='database')
        db_api.setup_db()

        enf = policy.init(use_conf=False)
        for default in enf.registered_rules.values():
            if default.name not in enf.rules:
                enf.rules[default.name] = default.check

        self.config(
            custom_artifact_types_modules=[
                'glare.tests.sample_artifact',
                'glare.tests.hooks_artifact',
                'glare.tests.unpacking_artifact'
            ],
            enabled_artifact_types=[
                'unpacking_artifact', 'hooks_artifact', 'sample_artifact',
                'images', 'heat_templates', 'heat_environments',
                'murano_packages', 'tosca_templates']
        )

        location.SCHEME_TO_CLS_MAP = {}
        self._create_stores()
        self.addCleanup(setattr, location, 'SCHEME_TO_CLS_MAP', dict())

        self.addCleanup(db_api.drop_db)
        self.addCleanup(policy.reset)

    def config(self, **kw):
        """Override some configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a group argument is supplied, the overrides are applied to
        the specified configuration option group.

        All overrides are automatically cleared at the end of the current
        test by the fixtures cleanup process.
        """
        self._config_fixture.config(**kw)

    @staticmethod
    def policy(new_rules):
        enf = policy.init(use_conf=False)
        for rule_name, rule_check_str in new_rules.items():
            enf.rules[rule_name] = os_policy.RuleDefault(
                rule_name, rule_check_str).check

    @staticmethod
    def get_fake_request(user):
        req = wsgi.Request.blank('')
        req.method = 'POST'
        kwargs = {
            'user': user['id'],
            'tenant': user['tenant_id'],
            'roles': user['roles'],
            'is_admin': 'admin' in user['roles'],
        }
        req.context = context.RequestContext(**kwargs)
        return req

    def _create_stores(self):
        """Create known stores. Mock out sheepdog's subprocess dependency
        on collie.

        :returns: the number of how many store drivers been loaded.
        """
        store.register_opts(CONF)

        self.config(default_store='filesystem',
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")

        store.create_stores(CONF)

    @staticmethod
    def generate_json_patch(values):
        patch = jsonpatch.JsonPatch(values)
        tuple(map(patch._get_operation, patch.patch))
        return patch

    def update_with_values(self, values, exc_class=None,
                           art_type='sample_artifact', art_id=None):
        patch = self.generate_json_patch(values)
        art_id = art_id or self.sample_artifact['id']
        if exc_class is None:
            return self.controller.update(self.req, art_type, art_id, patch)
        else:
            self.assertRaises(exc_class, self.controller.update, self.req,
                              art_type, art_id, patch)


class BaseTestArtifactAPI(BaseTestCase):

    def setUp(self):
        super(BaseTestArtifactAPI, self).setUp()
        self.controller = resource.ArtifactsController()
        self.req = self.get_fake_request(user=self.users['user1'])
        self.config(default_store='database',
                    group='artifact_type:sample_artifact')



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

"""
Guidelines for writing new hacking checks

 - Use only for Glare-specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range G3xx. Find the current test with
   the highest allocated number and then pick the next value.
   If nova has an N3xx code for that test, use the same number.
 - Keep the test method code in the source file ordered based
   on the G3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to glare/tests/test_hacking.py

"""


asse_trueinst_re = re.compile(
    r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
    "(\w|\.|\'|\"|\[|\])+\)\)")
asse_equal_type_re = re.compile(
    r"(.)*assertEqual\(type\((\w|\.|\'|\"|\[|\])+\), "
    "(\w|\.|\'|\"|\[|\])+\)")
asse_equal_end_with_none_re = re.compile(
    r"(.)*assertEqual\((\w|\.|\'|\"|\[|\])+, None\)")
asse_equal_start_with_none_re = re.compile(
    r"(.)*assertEqual\(None, (\w|\.|\'|\"|\[|\])+\)")
unicode_func_re = re.compile(r"(\s|\W|^)unicode\(")

_all_log_levels = {'debug', 'error', 'info', 'warning',
                   'critical', 'exception'}
# Since _Lx have been removed, we just need to check _()
translated_logs = re.compile(
    r"(.)*LOG\.(%(level)s)\(\s*_\(" % {'level': '|'.join(_all_log_levels)})

dict_constructor_with_list_copy_re = re.compile(r".*\bdict\((\[)?(\(|\[)")


def assert_true_instance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    G316
    """
    if asse_trueinst_re.match(logical_line):
        yield (0, "G316: assertTrue(isinstance(a, b)) sentences not allowed")


def assert_equal_type(logical_line):
    """Check for assertEqual(type(A), B) sentences

    G317
    """
    if asse_equal_type_re.match(logical_line):
        yield (0, "G317: assertEqual(type(A), B) sentences not allowed")


def assert_equal_none(logical_line):
    """Check for assertEqual(A, None) or assertEqual(None, A) sentences

    G318
    """
    res = (asse_equal_start_with_none_re.match(logical_line) or
           asse_equal_end_with_none_re.match(logical_line))
    if res:
        yield (0, "G318: assertEqual(A, None) or assertEqual(None, A) "
               "sentences not allowed")


def no_translate_logs(logical_line):
    """Check for use of LOG.*(_(

    G319
    """
    if translated_logs.match(logical_line):
        yield (0, "G319: Don't translate logs")


def no_direct_use_of_unicode_function(logical_line):
    """Check for use of unicode() builtin

    G320
    """
    if unicode_func_re.match(logical_line):
        yield(0, "G320: Use six.text_type() instead of unicode()")


def check_no_contextlib_nested(logical_line):
    msg = ("G327: contextlib.nested is deprecated since Python 2.7. See "
           "https://docs.python.org/2/library/contextlib.html#contextlib."
           "nested for more information.")
    if ("with contextlib.nested(" in logical_line or
            "with nested(" in logical_line):
        yield(0, msg)


def dict_constructor_with_list_copy(logical_line):
    msg = ("G328: Must use a dict comprehension instead of a dict constructor "
           "with a sequence of key-value pairs.")
    if dict_constructor_with_list_copy_re.match(logical_line):
        yield (0, msg)


def check_python3_xrange(logical_line):
    if re.search(r"\bxrange\s*\(", logical_line):
        yield(0, "G329: Do not use xrange. Use range, or six.moves.range for "
                 "large loops.")


def check_python3_no_iteritems(logical_line):
    msg = ("G330: Use six.iteritems() or dict.items() instead of "
           "dict.iteritems().")
    if re.search(r".*\.iteritems\(\)", logical_line):
        yield(0, msg)


def check_python3_no_iterkeys(logical_line):
    msg = ("G331: Use six.iterkeys() or dict.keys() instead of "
           "dict.iterkeys().")
    if re.search(r".*\.iterkeys\(\)", logical_line):
        yield(0, msg)


def check_python3_no_itervalues(logical_line):
    msg = ("G332: Use six.itervalues() or dict.values instead of "
           "dict.itervalues().")
    if re.search(r".*\.itervalues\(\)", logical_line):
        yield(0, msg)


def factory(register):
    register(assert_true_instance)
    register(assert_equal_type)
    register(assert_equal_none)
    register(no_translate_logs)
    register(no_direct_use_of_unicode_function)
    register(check_no_contextlib_nested)
    register(dict_constructor_with_list_copy)
    register(check_python3_xrange)
    register(check_python3_no_iteritems)
    register(check_python3_no_iterkeys)
    register(check_python3_no_itervalues)



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile

import mock
from OpenSSL import crypto
import six

from glare.common import exception as exc
from glare.common import utils
from glare.tests.unit import base


class TestUtils(base.BaseTestCase):
    """Test class for glare.common.utils"""

    def test_validate_quotes(self):
        self.assertIsNone(utils.validate_quotes('"classic"'))
        self.assertIsNone(utils.validate_quotes('This is a good string'))
        self.assertIsNone(utils.validate_quotes
                          ('"comma after quotation mark should work",'))
        self.assertIsNone(utils.validate_quotes
                          (',"comma before quotation mark should work"'))
        self.assertIsNone(utils.validate_quotes('"we have quotes \\" inside"'))

    def test_validate_quotes_negative(self):
        self.assertRaises(exc.InvalidParameterValue,
                          utils.validate_quotes, 'not_comma"blabla"')
        self.assertRaises(exc.InvalidParameterValue, utils.validate_quotes,
                          '"No comma after quotation mark"Not_comma')
        self.assertRaises(exc.InvalidParameterValue,
                          utils.validate_quotes, '"The quote is not closed')

    def test_no_4bytes_params(self):
        @utils.no_4byte_params
        def test_func(*args, **kwargs):
            return args, kwargs

        bad_char = u'\U0001f62a'

        # params without 4bytes unicode are okay
        args, kwargs = test_func('val1', param='val2')
        self.assertEqual(('val1',), args)
        self.assertEqual({'param': 'val2'}, kwargs)

        # test various combinations with bad param
        self.assertRaises(exc.BadRequest, test_func,
                          bad_char)
        self.assertRaises(exc.BadRequest, test_func,
                          **{bad_char: 'val1'})
        self.assertRaises(exc.BadRequest, test_func,
                          **{'param': bad_char})


class TestReaders(base.BaseTestCase):
    """Test various readers in glare.common.utils"""

    def test_cooperative_reader_iterator(self):
        """Ensure cooperative reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            for chunk in utils.CooperativeReader(tmp_fd):
                bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

    def test_cooperative_reader_explicit_read(self):
        BYTES = 1024
        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            reader = utils.CooperativeReader(tmp_fd)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_cooperative_reader_no_read_method(self):
        BYTES = 1024
        stream = [b'*'] * BYTES
        reader = utils.CooperativeReader(stream)
        bytes_read = 0
        byte = reader.read()
        while len(byte) != 0:
            bytes_read += 1
            byte = reader.read()

        self.assertEqual(BYTES, bytes_read)

        # some data may be left in the buffer
        reader = utils.CooperativeReader(stream)
        reader.buffer = 'some data'
        buffer_string = reader.read()
        self.assertEqual('some data', buffer_string)

    def test_cooperative_reader_no_read_method_buffer_size(self):
        # Decrease buffer size to 1000 bytes to test its overflow
        with mock.patch('glare.common.utils.MAX_COOP_READER_BUFFER_SIZE',
                        1000):
            BYTES = 1024
            stream = [b'*'] * BYTES
            reader = utils.CooperativeReader(stream)
            # Reading 1001 bytes to the buffer leads to 413 error
            self.assertRaises(exc.RequestEntityTooLarge, reader.read, 1001)

    def test_cooperative_reader_of_iterator(self):
        """Ensure cooperative reader supports iterator backends too"""
        data = b'abcdefgh'
        data_list = [data[i:i + 1] * 3 for i in range(len(data))]
        reader = utils.CooperativeReader(data_list)
        chunks = []
        while True:
            chunks.append(reader.read(3))
            if chunks[-1] == b'':
                break
        meat = b''.join(chunks)
        self.assertEqual(b'aaabbbcccdddeeefffggghhh', meat)

    def test_cooperative_reader_of_iterator_stop_iteration_err(self):
        """Ensure cooperative reader supports iterator backends too"""
        reader = utils.CooperativeReader([l * 3 for l in ''])
        chunks = []
        while True:
            chunks.append(reader.read(3))
            if chunks[-1] == b'':
                break
        meat = b''.join(chunks)
        self.assertEqual(b'', meat)

    def _create_generator(self, chunk_size, max_iterations):
        chars = b'abc'
        iteration = 0
        while True:
            index = iteration % len(chars)
            chunk = chars[index:index + 1] * chunk_size
            yield chunk
            iteration += 1
            if iteration >= max_iterations:
                raise StopIteration()

    def _test_reader_chunked(self, chunk_size, read_size, max_iterations=5):
        generator = self._create_generator(chunk_size, max_iterations)
        reader = utils.CooperativeReader(generator)
        result = bytearray()
        while True:
            data = reader.read(read_size)
            if len(data) == 0:
                break
            self.assertLessEqual(len(data), read_size)
            result += data
        expected = (b'a' * chunk_size +
                    b'b' * chunk_size +
                    b'c' * chunk_size +
                    b'a' * chunk_size +
                    b'b' * chunk_size)
        self.assertEqual(expected, bytes(result))

    def test_cooperative_reader_preserves_size_chunk_less_then_read(self):
        self._test_reader_chunked(43, 101)

    def test_cooperative_reader_preserves_size_chunk_equals_read(self):
        self._test_reader_chunked(1024, 1024)

    def test_cooperative_reader_preserves_size_chunk_more_then_read(self):
        chunk_size = 16 * 1024 * 1024  # 16 Mb, as in remote http source
        read_size = 8 * 1024           # 8k, as in httplib
        self._test_reader_chunked(chunk_size, read_size)

    def test_limiting_reader(self):
        """Ensure limiting reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        data = six.BytesIO(b"*" * BYTES)
        for chunk in utils.LimitingReader(data, BYTES):
            bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

        bytes_read = 0
        data = six.BytesIO(b"*" * BYTES)
        reader = utils.LimitingReader(data, BYTES)
        byte = reader.read(1)
        while len(byte) != 0:
            bytes_read += 1
            byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_limiting_reader_fails(self):
        """Ensure limiting reader class throws exceptions if limit exceeded"""
        BYTES = 1024

        def _consume_all_iter():
            bytes_read = 0
            data = six.BytesIO(b"*" * BYTES)
            for chunk in utils.LimitingReader(data, BYTES - 1):
                bytes_read += len(chunk)

        self.assertRaises(exc.RequestEntityTooLarge, _consume_all_iter)

        def _consume_all_read():
            bytes_read = 0
            data = six.BytesIO(b"*" * BYTES)
            reader = utils.LimitingReader(data, BYTES - 1)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertRaises(exc.RequestEntityTooLarge, _consume_all_read)

    def test_blob_iterator(self):
        BYTES = 1024
        bytes_read = 0
        stream = [b'*'] * BYTES
        for chunk in utils.BlobIterator(stream, 64):
            bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)


class TestKeyCert(base.BaseTestCase):

    def test_validate_key_cert_key(self):
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../', 'var'))
        keyfile = os.path.join(var_dir, 'privatekey.key')
        certfile = os.path.join(var_dir, 'certificate.crt')
        utils.validate_key_cert(keyfile, certfile)

    def test_validate_key_cert_no_private_key(self):
        with tempfile.NamedTemporaryFile('w+') as tmpf:
            self.assertRaises(RuntimeError,
                              utils.validate_key_cert,
                              "/not/a/file", tmpf.name)

    def test_validate_key_cert_cert_cant_read(self):
        with tempfile.NamedTemporaryFile('w+') as keyf:
            with tempfile.NamedTemporaryFile('w+') as certf:
                os.chmod(certf.name, 0)
                self.assertRaises(RuntimeError,
                                  utils.validate_key_cert,
                                  keyf.name, certf.name)

    def test_validate_key_cert_key_cant_read(self):
        with tempfile.NamedTemporaryFile('w+') as keyf:
            with tempfile.NamedTemporaryFile('w+') as certf:
                os.chmod(keyf.name, 0)
                self.assertRaises(RuntimeError,
                                  utils.validate_key_cert,
                                  keyf.name, certf.name)

    def test_validate_key_cert_key_crypto_error(self):
        var_dir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                               '../', 'var'))
        keyfile = os.path.join(var_dir, 'privatekey.key')
        certfile = os.path.join(var_dir, 'certificate.crt')
        with mock.patch('OpenSSL.crypto.verify', side_effect=crypto.Error):
            self.assertRaises(RuntimeError,
                              utils.validate_key_cert,
                              keyfile, certfile)



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class LockApiBase(object):
    """Lock Api Base class that responsible for acquiring/releasing locks."""

    def create_lock(self, context, lock_key):
        """Acquire lock for current user.

        :param context: user context
        :param lock_key: unique lock identifier that defines lock scope
        :return: lock internal identifier
        """
        raise NotImplementedError()

    def delete_lock(self, context, lock_id):
        """Delete acquired user lock.

        :param context: user context
        :param lock_id: lock internal identifier
        """
        raise NotImplementedError()


class Lock(object):
    """Object that stores lock context for users. This class is internal
    and used only in lock engine, so users shouldn't use this class directly.
    """

    def __init__(self, context, lock_id, lock_key, release_method):
        """Initialize lock context."""
        self.context = context
        self.lock_id = lock_id
        self.lock_key = lock_key
        self.release = release_method

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # TODO(kairat) catch all exceptions here
        self.release(self)


class LockEngine(object):
    """Glare lock engine.

    Defines how artifact updates must be synchronized with each other. When
    some user obtains a lock for the same artifact then other user cannot
    request that lock and gets a Conflict error.
    """
    # NOTE(kairat): Lock Engine also allows to encapsulate lock logic in one
    # place so we can potentially add tooz functionality in future to Glare.
    # Right now there are troubles with locks in Galera (especially in mysql)
    # and zookeeper requires additional work from IT engineers. So we need
    # support production ready DB locks in our implementation.

    MAX_LOCK_LENGTH = 255

    def __init__(self, lock_api):
        """Initialize lock engine with some lock api.

        :param lock_api: api that allows to create/delete locks
        """
        # NOTE(kairat): lock_api is db_api now but it might be
        # replaced with DLM in near future.
        self.lock_api = lock_api

    def acquire(self, context, lock_key):
        """Acquire lock for artifact.

        If there is some other lock with the same key then
        raise Conflict Error.

        :param context: user context
        :param lock_key: lock key
        :return: lock definition
        """
        if lock_key is not None and len(lock_key) < self.MAX_LOCK_LENGTH:
            lock_id = self.lock_api.create_lock(context, lock_key)
            LOG.debug("Lock %(lock_id)s acquired for lock_key %(lock_key)s",
                      {'lock_id': lock_id, 'lock_key': lock_key})
        else:
            lock_id = None
            LOG.debug("No lock for lock_key %s", lock_key)

        return Lock(context, lock_id, lock_key, self.release)

    def release(self, lock):
        """Release lock for artifact.

        :param lock: Lock object
        """
        if lock.lock_id is not None:
            self.lock_api.delete_lock(lock.context, lock.lock_id)
            LOG.debug("Lock %(lock_id)s released for lock_key %(key)s",
                      {'lock_id': lock.lock_id, 'key': lock.lock_key})



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import operator
import threading

from oslo_config import cfg
from oslo_db import exception as db_exception
from oslo_db import options
from oslo_db.sqlalchemy import session
from oslo_log import log as os_logging
from oslo_utils import timeutils
import osprofiler.sqlalchemy
from retrying import retry
import six
import sqlalchemy
from sqlalchemy import and_
import sqlalchemy.exc
from sqlalchemy import exists
from sqlalchemy import func
from sqlalchemy import or_
import sqlalchemy.orm as orm
from sqlalchemy.orm import aliased
from sqlalchemy.orm import joinedload

from glare.common import exception
from glare.common import semver_db
from glare.common import utils
from glare.db.sqlalchemy import models
from glare.i18n import _

LOG = os_logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_group("profiler", "glare.common.wsgi")
options.set_defaults(CONF)


BASE_ARTIFACT_PROPERTIES = ('id', 'visibility', 'created_at', 'updated_at',
                            'activated_at', 'owner', 'status', 'description',
                            'name', 'type_name', 'version')

_FACADE = None
_LOCK = threading.Lock()


def _retry_on_deadlock(exc):
    """Decorator to retry a DB API call if Deadlock was received."""

    if isinstance(exc, db_exception.DBDeadlock):
        LOG.warning("Deadlock detected. Retrying...")
        return True
    return False


def _create_facade_lazily():
    global _LOCK, _FACADE
    if _FACADE is None:
        with _LOCK:
            if _FACADE is None:
                _FACADE = session.EngineFacade.from_config(CONF)

                if CONF.profiler.enabled and CONF.profiler.trace_sqlalchemy:
                    osprofiler.sqlalchemy.add_tracing(sqlalchemy,
                                                      _FACADE.get_engine(),
                                                      "db")
    return _FACADE


def get_engine():
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(autocommit=True, expire_on_commit=False):
    facade = _create_facade_lazily()
    return facade.get_session(autocommit=autocommit,
                              expire_on_commit=expire_on_commit)


def setup_db():
    engine = get_engine()
    models.register_models(engine)


def drop_db():
    engine = get_engine()
    models.unregister_models(engine)


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def delete(context, artifact_id, session):
    with session.begin():
        session.query(models.Artifact).filter_by(id=artifact_id).delete()


def _drop_protected_attrs(model_class, values):
    """Removed protected attributes from values dictionary using the models
    __protected_attributes__ field.
    """
    for attr in model_class.__protected_attributes__:
        if attr in values:
            del values[attr]


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
@utils.no_4byte_params
def create_or_update(context, artifact_id, values, session):
    with session.begin():
        _drop_protected_attrs(models.Artifact, values)
        if artifact_id is None:
            # create new artifact
            artifact = models.Artifact()
            artifact.id = values.pop('id')
        else:
            # update the existing artifact
            artifact = _get(context, artifact_id, session)

        if 'version' in values:
            values['version'] = semver_db.parse(values['version'])

        if 'tags' in values:
            tags = values.pop('tags')
            artifact.tags = _do_tags(artifact, tags)

        if 'properties' in values:
            properties = values.pop('properties', {})
            artifact.properties = _do_properties(artifact, properties)

        if 'blobs' in values:
            blobs = values.pop('blobs')
            artifact.blobs = _do_blobs(artifact, blobs)

        artifact.updated_at = timeutils.utcnow()
        if 'status' in values:
            if session.query(exists().where(and_(
                models.ArtifactBlob.status == 'saving',
                models.ArtifactBlob.artifact_id == artifact_id))
            ).one()[0]:
                raise exception.Conflict(
                    "You cannot change artifact status if it has "
                    "uploading blobs.")
            if values['status'] == 'active':
                artifact.activated_at = timeutils.utcnow()
        artifact.update(values)

        artifact.save(session=session)
        LOG.debug("Response from the database was received.")

        return artifact.to_dict()


def _get(context, artifact_id, session):
    try:
        query = _do_artifacts_query(context, session).filter_by(
            id=artifact_id)
        artifact = query.one()
    except orm.exc.NoResultFound:
        msg = _("Artifact with id=%s not found.") % artifact_id
        LOG.warning(msg)
        raise exception.ArtifactNotFound(msg)
    return artifact


def get(context, artifact_id, session):
    return _get(context, artifact_id, session).to_dict()


def get_all(context, session, filters=None, marker=None, limit=None,
            sort=None, latest=False):
    """List all visible artifacts

    :param filters: dict of filter keys and values.
    :param marker: artifact id after which to start page
    :param limit: maximum number of artifacts to return
    :param sort: a tuple (key, dir, type) where key is an attribute by
     which results should be sorted, dir is a direction: 'asc' or 'desc',
     and type is type of the attribute: 'bool', 'string', 'numeric' or 'int' or
     None if attribute is base.
    :param latest: flag that indicates, that only artifacts with highest
     versions should be returned in output
    """
    artifacts = _get_all(
        context, session, filters, marker, limit, sort, latest)
    return [af.to_dict() for af in artifacts]


def _apply_latest_filter(context, session, query,
                         basic_conds, tag_conds, prop_conds):
    # Subquery to fetch max version suffix for a group (name,
    # version_prefix)
    ver_suffix_subq = _apply_query_base_filters(
        session.query(
            models.Artifact.name,
            models.Artifact.version_prefix,
            func.max(models.Artifact.version_suffix).label(
                'max_suffix')).group_by(
            models.Artifact.name, models.Artifact.version_prefix),
        context)
    ver_suffix_subq = _apply_user_filters(
        ver_suffix_subq, basic_conds, tag_conds, prop_conds).subquery()
    # Subquery to fetch max version prefix for a name group
    ver_prefix_subq = _apply_query_base_filters(
        session.query(models.Artifact.name, func.max(
            models.Artifact.version_prefix).label('max_prefix')).group_by(
            models.Artifact.name),
        context)
    ver_prefix_subq = _apply_user_filters(
        ver_prefix_subq, basic_conds, tag_conds, prop_conds).subquery()
    # Combine two subqueries together joining them with Artifact table
    query = query.join(
        ver_prefix_subq,
        and_(models.Artifact.name == ver_prefix_subq.c.name,
             models.Artifact.version_prefix ==
             ver_prefix_subq.c.max_prefix)).join(
        ver_suffix_subq,
        and_(models.Artifact.name == ver_suffix_subq.c.name,
             models.Artifact.version_prefix ==
             ver_suffix_subq.c.version_prefix,
             models.Artifact.version_suffix ==
             ver_suffix_subq.c.max_suffix)
    )

    return query


def _apply_user_filters(query, basic_conds, tag_conds, prop_conds):

    if basic_conds:
        for basic_condition in basic_conds:
            query = query.filter(and_(*basic_condition))

    if tag_conds:
        for tag_condition in tag_conds:
            query = query.join(models.ArtifactTag, aliased=True).filter(
                and_(*tag_condition))

    if prop_conds:
        for prop_condition in prop_conds:
            query = query.join(models.ArtifactProperty, aliased=True).filter(
                and_(*prop_condition))

    return query


def _get_all(context, session, filters=None, marker=None, limit=None,
             sort=None, latest=False):

    filters = filters or {}

    query = _do_artifacts_query(context, session)

    basic_conds, tag_conds, prop_conds = _do_query_filters(filters)

    query = _apply_user_filters(query, basic_conds, tag_conds, prop_conds)

    if latest:
        query = _apply_latest_filter(context, session, query,
                                     basic_conds, tag_conds, prop_conds)

    marker_artifact = None
    if marker is not None:
        marker_artifact = get(context, marker, session)

    query = _do_paginate_query(query=query, limit=limit,
                               marker=marker_artifact, sort=sort)

    return query.all()


def _do_paginate_query(query, marker=None, limit=None, sort=None):
    # Add sorting
    number_of_custom_props = 0
    for sort_key, sort_dir, sort_type in sort:
        try:
            sort_dir_func = {
                'asc': sqlalchemy.asc,
                'desc': sqlalchemy.desc,
            }[sort_dir]
        except KeyError:
            msg = _("Unknown sort direction, must be 'desc' or 'asc'.")
            raise exception.BadRequest(msg)
        # Note(mfedosin): Workaround to deal with situation that sqlalchemy
        # cannot work with composite keys correctly
        if sort_key == 'version':
            query = query.order_by(sort_dir_func(models.Artifact.version_prefix))\
                         .order_by(sort_dir_func(models.Artifact.version_suffix))\
                         .order_by(sort_dir_func(models.Artifact.version_meta))
        elif sort_key in BASE_ARTIFACT_PROPERTIES:
            # sort by generic property
            query = query.order_by(sort_dir_func(getattr(models.Artifact,
                                                         sort_key)))
        else:
            # sort by custom property
            number_of_custom_props += 1
            if number_of_custom_props > 1:
                msg = _("For performance sake it's not allowed to sort by "
                        "more than one custom property with this db backend.")
                raise exception.BadRequest(msg)
            prop_table = aliased(models.ArtifactProperty)
            query = (
                query.join(prop_table).
                filter(prop_table.name == sort_key).
                order_by(sort_dir_func(getattr(prop_table,
                                               sort_type + '_value'))))

    # Add pagination
    if marker is not None:
        marker_values = []
        for sort_key, __, __ in sort:
            v = marker.get(sort_key, None)
            marker_values.append(v)

        # Build up an array of sort criteria as in the docstring
        criteria_list = []
        for i in range(len(sort)):
            crit_attrs = []
            for j in range(i):
                value = marker_values[j]
                if sort[j][0] in BASE_ARTIFACT_PROPERTIES:
                    if sort[j][0] == 'version':
                        value = semver_db.parse(value)
                    crit_attrs.append([getattr(models.Artifact, sort[j][0]) ==
                                       value])
                else:
                    conds = [models.ArtifactProperty.name == sort[j][0]]
                    conds.extend([getattr(models.ArtifactProperty,
                                 sort[j][2] + '_value') == value])
                    crit_attrs.append(conds)

            value = marker_values[i]
            sort_dir_func = operator.gt if sort[i][1] == 'asc' else operator.lt
            if sort[i][0] in BASE_ARTIFACT_PROPERTIES:
                if sort[i][0] == 'version':
                    value = semver_db.parse(value)
                crit_attrs.append([sort_dir_func(getattr(models.Artifact,
                                                         sort[i][0]), value)])
            else:
                query = query.join(models.ArtifactProperty, aliased=True)
                conds = [models.ArtifactProperty.name == sort[i][0]]
                conds.extend([sort_dir_func(getattr(models.ArtifactProperty,
                             sort[i][2] + '_value'), value)])
                crit_attrs.append(conds)

            criteria = [and_(*crit_attr) for crit_attr in crit_attrs]
            criteria_list.append(criteria)

        criteria_list = [and_(*cr) for cr in criteria_list]
        query = query.filter(or_(*criteria_list))

    if limit is not None:
        query = query.limit(limit)

    return query


def _do_artifacts_query(context, session):
    """Build the query to get all artifacts based on the context"""

    query = session.query(models.Artifact)

    query = (query.options(joinedload(models.Artifact.properties)).
             options(joinedload(models.Artifact.tags)).
             options(joinedload(models.Artifact.blobs)))

    return _apply_query_base_filters(query, context)


def _apply_query_base_filters(query, context):
    # If admin, return everything.
    if context.is_admin:
        return query

    # If anonymous user, return only public artifacts.
    # However, if context.tenant has a value, return both
    # public and private artifacts of the owner.
    if context.tenant is not None:
        query = query.filter(
            or_(models.Artifact.owner == context.tenant,
                models.Artifact.visibility == 'public'))
    else:
        query = query.filter(
            models.Artifact.visibility == 'public')

    return query

op_mappings = {
    'eq': operator.eq,
    'gt': operator.gt,
    'gte': operator.ge,
    'lt': operator.lt,
    'lte': operator.le,
    'neq': operator.ne,
}


def _do_query_filters(filters):
    basic_conds = []
    tag_conds = []
    prop_conds = []
    for field_name, key_name, op, field_type, value in filters:
        if field_name == 'tags':
            tags = utils.split_filter_value_for_quotes(value)
            for tag in tags:
                tag_conds.append([models.ArtifactTag.value == tag])
        elif field_name == 'tags-any':
            tags = utils.split_filter_value_for_quotes(value)
            tag_conds.append([models.ArtifactTag.value.in_(tags)])
        elif field_name in BASE_ARTIFACT_PROPERTIES:
            if op != 'in':
                fn = op_mappings[op]
                if field_name == 'version':
                    value = semver_db.parse(value)
                basic_conds.append([fn(getattr(models.Artifact, field_name),
                                       value)])
            else:
                if field_name == 'version':
                    value = [semver_db.parse(val) for val in value]
                    basic_conds.append(
                        [or_(*[
                            models.Artifact.version == ver for ver in value])])
                else:
                    basic_conds.append(
                        [getattr(models.Artifact, field_name).in_(value)])
        else:
            conds = [models.ArtifactProperty.name == field_name]
            if key_name is not None:
                if op == 'eq' or value is not None:
                    conds.extend(
                        [models.ArtifactProperty.key_name == key_name])
                elif op == 'in':
                    conds.extend(
                        [models.ArtifactProperty.key_name.in_(key_name)])
            if value is not None:
                if op != 'in':
                    fn = op_mappings[op]
                    conds.extend([fn(getattr(models.ArtifactProperty,
                                             field_type + '_value'), value)])
                else:
                    conds.extend([getattr(models.ArtifactProperty,
                                          field_type + '_value').in_(value)])

            prop_conds.append(conds)

    return basic_conds, tag_conds, prop_conds


def _do_tags(artifact, new_tags):
    tags_to_update = []
    # don't touch existing tags
    for tag in artifact.tags:
        if tag.value in new_tags:
            tags_to_update.append(tag)
            new_tags.remove(tag.value)
    # add new tags
    for tag in new_tags:
        db_tag = models.ArtifactTag()
        db_tag.value = tag
        tags_to_update.append(db_tag)
    return tags_to_update


def _get_prop_type(value):
    if isinstance(value, bool):
        return 'bool_value'
    if isinstance(value, int):
        return 'int_value'
    if isinstance(value, six.string_types):
        return 'string_value'
    if isinstance(value, float):
        return 'numeric_value'


def _create_property(prop_name, prop_value, position=None, key_name=None):
    db_prop = models.ArtifactProperty()
    db_prop.name = prop_name
    setattr(db_prop, _get_prop_type(prop_value), prop_value)
    db_prop.position = position
    db_prop.key_name = key_name
    return db_prop


def _do_properties(artifact, new_properties):
    props_to_update = []
    # don't touch the existing properties
    for prop in artifact.properties:
        if prop.name not in new_properties:
            props_to_update.append(prop)

    for prop_name, prop_value in new_properties.items():
        if prop_value is None:
            continue
        if isinstance(prop_value, list):
            for pos, list_prop in enumerate(prop_value):
                for prop in artifact.properties:
                    if prop.name == prop_name and pos == prop.position:
                        if getattr(prop, _get_prop_type(
                                list_prop)) != list_prop:
                            setattr(prop, _get_prop_type(list_prop),
                                    list_prop)
                        props_to_update.append(prop)
                        break
                else:
                    props_to_update.append(
                        _create_property(prop_name, list_prop, position=pos)
                    )
        elif isinstance(prop_value, dict):
            for dict_key, dict_val in prop_value.items():
                for prop in artifact.properties:
                    if prop.name == prop_name and prop.key_name == dict_key:
                        if getattr(prop, _get_prop_type(dict_val)) != dict_val:
                            setattr(prop, _get_prop_type(dict_val), dict_val)
                        props_to_update.append(prop)
                        break
                else:
                    props_to_update.append(
                        _create_property(prop_name, dict_val,
                                         key_name=dict_key)
                    )
        elif prop_value is not None:
            for prop in artifact.properties:
                if prop.name == prop_name:
                    setattr(prop, _get_prop_type(prop_value), prop_value)
                    props_to_update.append(prop)
                    break
            else:
                props_to_update.append(_create_property(
                    prop_name, prop_value))

    return props_to_update


def _update_blob_values(blob, values):
    for elem in ('size', 'md5', 'sha1', 'sha256', 'url', 'external', 'status',
                 'content_type'):
        setattr(blob, elem, values[elem])
    return blob


def _do_blobs(artifact, new_blobs):
    blobs_to_update = []
    # don't touch the existing blobs
    for blob in artifact.blobs:
        if blob.name not in new_blobs:
            blobs_to_update.append(blob)

    for blob_name, blob_value in new_blobs.items():
        if blob_value is None:
            continue
        if isinstance(blob_value.get('status'), str):
            for blob in artifact.blobs:
                if blob.name == blob_name:
                    _update_blob_values(blob, blob_value)
                    blobs_to_update.append(blob)
                    break
            else:
                blob = models.ArtifactBlob()
                blob.name = blob_name
                _update_blob_values(blob, blob_value)
                blobs_to_update.append(blob)
        else:
            for dict_key, dict_val in blob_value.items():
                for blob in artifact.blobs:
                    if blob.name == blob_name and blob.key_name == dict_key:
                        _update_blob_values(blob, dict_val)
                        blobs_to_update.append(blob)
                        break
                else:
                    blob = models.ArtifactBlob()
                    blob.name = blob_name
                    blob.key_name = dict_key
                    _update_blob_values(blob, dict_val)
                    blobs_to_update.append(blob)

    return blobs_to_update


def count_artifact_number(context, session, type_name=None):
    """Return a number of artifacts for tenant."""
    query = session.query(func.count(models.Artifact.id)).filter(
        models.Artifact.owner == context.tenant)
    if type_name is not None:
        query = query.filter(models.Artifact.type_name == type_name)
    return query.order_by(None).scalar() or 0


def calculate_uploaded_data(context, session, type_name=None):
    """Return the amount of uploaded data for tenant."""
    query = session.query(
        func.sum(models.ArtifactBlob.size)).join(
        models.Artifact, aliased=True).filter(
        models.Artifact.owner == context.tenant)
    if type_name is not None:
        query = query.filter(models.Artifact.type_name == type_name)
    return query.order_by(None).scalar() or 0


def _generate_quota_id(project_id, quota_name, type_name=None):
    quota_id = b"%s:%s" % (project_id.encode(), quota_name.encode())
    if type_name is not None:
        quota_id += b":%s" % type_name.encode()
    return hashlib.md5(quota_id).hexdigest()


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
@utils.no_4byte_params
def set_quotas(values, session):
    """Create new quota instances in database"""
    with session.begin():
        for project_id, project_quotas in values.items():

            # reset all project quotas
            session.query(models.ArtifactQuota).filter(
                models.ArtifactQuota.project_id == project_id).delete()

            # generate new quotas
            for quota_name, quota_value in project_quotas.items():
                q = models.ArtifactQuota()
                q.project_id = project_id
                q.quota_name = quota_name
                q.quota_value = quota_value
                session.add(q)

        # save all quotas
        session.flush()


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def get_all_quotas(session, project_id=None):
    """List all available quotas."""
    query = session.query(models.ArtifactQuota)
    if project_id is not None:
        query = query.filter(
            models.ArtifactQuota.project_id == project_id)
    quotas = query.order_by(models.ArtifactQuota.project_id).all()

    res = {}
    for quota in quotas:
        res.setdefault(
            quota.project_id, {})[quota.quota_name] = quota.quota_value

    return res


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
@utils.no_4byte_params
def create_lock(context, lock_key, session):
    """Try to create lock record."""
    with session.begin():
        existing = session.query(models.ArtifactLock).get(lock_key)
        if existing is None:
            try:
                lock = models.ArtifactLock()
                lock.id = lock_key
                lock.save(session=session)
                return lock.id
            except (sqlalchemy.exc.IntegrityError,
                    db_exception.DBDuplicateEntry):
                msg = _("Cannot lock an item with key %s. "
                        "Lock already acquired by other request") % lock_key
                raise exception.Conflict(msg)
        else:
            if timeutils.is_older_than(existing.acquired_at, 5):
                existing.acquired_at = timeutils.utcnow()
                existing.save(session)
                return existing.id
            else:
                msg = _("Cannot lock an item with key %s. "
                        "Lock already acquired by other request") % lock_key
                raise exception.Conflict(msg)


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def delete_lock(context, lock_id, session):
    with session.begin():
        session.query(models.ArtifactLock).filter_by(id=lock_id).delete()


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def save_blob_data(context, blob_data_id, data, session):
    """Save blob data to database."""
    with session.begin():
        blob_data = models.ArtifactBlobData()
        blob_data.id = blob_data_id
        blob_data.data = data.read()
        blob_data.save(session=session)
        return "sql://" + blob_data.id


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def save_blob_data_batch(context, blobs, session):
    """Perform batch uploading to database."""
    with session.begin():

        locations = []

        # blobs is a list of tuples (blob_data_id, data)
        for blob_data_id, data in blobs:
            blob_data = models.ArtifactBlobData()
            blob_data.id = blob_data_id
            blob_data.data = data.read()
            session.add(blob_data)
            locations.append("sql://" + blob_data.id)

        session.flush()

    return locations


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def get_blob_data(context, uri, session):
    """Download blob data from database."""

    blob_data_id = uri[6:]
    try:
        blob_data = session.query(
            models.ArtifactBlobData).filter_by(id=blob_data_id).one()
    except orm.exc.NoResultFound:
        msg = _("Cannot find a blob data with id %s.") % blob_data_id
        raise exception.NotFound(msg)
    return blob_data.data


@retry(retry_on_exception=_retry_on_deadlock, wait_fixed=500,
       stop_max_attempt_number=50)
def delete_blob_data(context, uri, session):
    """Delete blob data from database."""
    with session.begin():
        blob_data_id = uri[6:]
        session.query(
            models.ArtifactBlobData).filter_by(id=blob_data_id).delete()
